# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from attr import attrib, attrs

from datumaro.components.annotation import Bbox, Label
from datumaro.components.errors import FailedLabelVotingError
from datumaro.util.annotation_util import mean_bbox, segment_iou

from .matcher import (
    AnnotationMatcher,
    BboxMatcher,
    CaptionsMatcher,
    Cuboid2DMatcher,
    Cuboid3dMatcher,
    FeatureVectorMatcher,
    HashKeyMatcher,
    ImageAnnotationMatcher,
    LabelMatcher,
    LineMatcher,
    MaskMatcher,
    PointsMatcher,
    PolygonMatcher,
    RotatedBboxMatcher,
    ShapeMatcher,
    TabularMatcher,
)

__all__ = [
    "AnnotationMerger",
    "LabelMerger",
    "BboxMerger",
    "RotatedBboxMerger",
    "PolygonMerger",
    "MaskMerger",
    "PointsMerger",
    "LineMerger",
    "CaptionsMerger",
    "Cuboid3dMerger",
    "ImageAnnotationMerger",
    "EllipseMerger",
    "HashKeyMerger",
    "FeatureVectorMerger",
]


@attrs(kw_only=True)
class AnnotationMerger(AnnotationMatcher):
    def merge_clusters(self, clusters):
        raise NotImplementedError()


@attrs(kw_only=True)
class LabelMerger(AnnotationMerger, LabelMatcher):
    quorum = attrib(converter=int, default=0)

    def merge_clusters(self, clusters):
        assert len(clusters) <= 1
        if len(clusters) == 0:
            return []

        votes = {}  # label -> score
        for ann in clusters[0]:
            label = self._context._get_src_label_name(ann, ann.label)
            votes[label] = 1 + votes.get(label, 0)

        merged = []
        for label, count in votes.items():
            if count < self.quorum:
                sources = set(
                    self.get_ann_source(id(a))
                    for a in clusters[0]
                    if label not in [self._context._get_src_label_name(l, l.label) for l in a]
                )
                sources = [self._context._dataset_map[s][1] for s in sources]
                self._context.add_item_error(FailedLabelVotingError, votes, sources=sources)
                continue

            merged.append(
                Label(
                    self._context._get_label_id(label),
                    attributes={"score": count / len(self._context._dataset_map)},
                )
            )

        return merged


@attrs(kw_only=True)
class _ShapeMerger(AnnotationMerger, ShapeMatcher):
    quorum = attrib(converter=int, default=0)

    def merge_clusters(self, clusters):
        return list(map(self.merge_cluster, clusters))

    def find_cluster_label(self, cluster):
        votes = {}
        for s in cluster:
            label = self._context._get_src_label_name(s, s.label)
            state = votes.setdefault(label, [0, 0])
            state[0] += s.attributes.get("score", 1.0)
            state[1] += 1

        label, (score, count) = max(votes.items(), key=lambda e: e[1][0])
        if count < self.quorum:
            self._context.add_item_error(FailedLabelVotingError, votes)
            label = None
        score = score / len(self._context._dataset_map)
        label = self._context._get_label_id(label)
        return label, score

    @staticmethod
    def _merge_cluster_shape_mean_box_nearest(cluster):
        mbbox = Bbox(*mean_bbox(cluster))
        dist = (segment_iou(mbbox, s) for s in cluster)
        nearest_pos, _ = max(enumerate(dist), key=lambda e: e[1])
        return cluster[nearest_pos]

    def merge_cluster_shape(self, cluster):
        shape = self._merge_cluster_shape_mean_box_nearest(cluster)
        shape_score = sum(max(0, self.distance(shape, s)) for s in cluster) / len(cluster)
        return shape, shape_score

    def merge_cluster(self, cluster):
        label, label_score = self.find_cluster_label(cluster)
        shape, shape_score = self.merge_cluster_shape(cluster)

        shape.z_order = max(cluster, key=lambda a: a.z_order).z_order
        shape.label = label
        shape.attributes["score"] = label_score * shape_score if label is not None else shape_score

        return shape


@attrs
class BboxMerger(_ShapeMerger, BboxMatcher):
    pass


@attrs
class PolygonMerger(_ShapeMerger, PolygonMatcher):
    pass


@attrs
class MaskMerger(_ShapeMerger, MaskMatcher):
    pass


@attrs
class PointsMerger(_ShapeMerger, PointsMatcher):
    pass


@attrs
class LineMerger(_ShapeMerger, LineMatcher):
    pass


@attrs
class CaptionsMerger(AnnotationMerger, CaptionsMatcher):
    pass


@attrs
class Cuboid3dMerger(_ShapeMerger, Cuboid3dMatcher):
    @staticmethod
    def _merge_cluster_shape_mean_box_nearest(cluster):
        raise NotImplementedError()
        # mbbox = Bbox(*mean_cuboid(cluster))
        # dist = (segment_iou(mbbox, s) for s in cluster)
        # nearest_pos, _ = max(enumerate(dist), key=lambda e: e[1])
        # return cluster[nearest_pos]

    def merge_cluster(self, cluster):
        label, label_score = self.find_cluster_label(cluster)
        shape, shape_score = self.merge_cluster_shape(cluster)

        shape.label = label
        shape.attributes["score"] = label_score * shape_score if label is not None else shape_score

        return shape


@attrs
class ImageAnnotationMerger(AnnotationMerger, ImageAnnotationMatcher):
    pass


@attrs
class EllipseMerger(_ShapeMerger, ShapeMatcher):
    pass


@attrs
class HashKeyMerger(AnnotationMerger, HashKeyMatcher):
    pass


@attrs
class FeatureVectorMerger(AnnotationMerger, FeatureVectorMatcher):
    pass


@attrs
class TabularMerger(AnnotationMerger, TabularMatcher):
    pass


@attrs
class RotatedBboxMerger(_ShapeMerger, RotatedBboxMatcher):
    pass


@attrs
class Cuboid2DMerger(_ShapeMerger, Cuboid2DMatcher):
    pass



# Copyright (C) 2019-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import logging as log
from typing import TYPE_CHECKING, Callable, Optional

# Disable B410: import_lxml - the library is used for writing
from lxml import etree as ET  # nosec

from datumaro.components.annotation import (
    Annotation,
    AnnotationType,
    Bbox,
    Caption,
    Ellipse,
    HashKey,
    Label,
    Mask,
    Points,
    Polygon,
    PolyLine,
)
from datumaro.components.media import Image
from datumaro.components.transformer import ItemTransform

if TYPE_CHECKING:
    from datumaro.components.dataset_base import CategoriesInfo, DatasetItem, IDataset

__all__ = [
    "XPathDatasetFilter",
    "XPathAnnotationsFilter",
    "UserFunctionDatasetFilter",
    "UserFunctionAnnotationsFilter",
]


class DatasetItemEncoder:
    @classmethod
    def encode(
        cls, item: DatasetItem, categories: Optional[CategoriesInfo] = None
    ) -> ET.ElementBase:
        item_elem = ET.Element("item")
        ET.SubElement(item_elem, "id").text = str(item.id)
        ET.SubElement(item_elem, "subset").text = str(item.subset)

        image = item.media
        if isinstance(image, Image):
            item_elem.append(cls.encode_image(image))

        for ann in item.annotations:
            item_elem.append(cls.encode_annotation(ann, categories))

        return item_elem

    @classmethod
    def encode_image(cls, image: Image) -> ET.ElementBase:
        image_elem = ET.Element("image")

        size = image.size
        if size is not None:
            h, w = str(size[0]), str(size[1])
        else:
            h = "unknown"
            w = h
        ET.SubElement(image_elem, "height").text = h
        ET.SubElement(image_elem, "width").text = w

        ET.SubElement(image_elem, "has_data").text = "%d" % int(image.has_data)
        if hasattr(image, "path"):
            ET.SubElement(image_elem, "path").text = image.path

        return image_elem

    @classmethod
    def encode_annotation_base(cls, annotation: Annotation) -> ET.ElementBase:
        assert isinstance(annotation, Annotation)
        ann_elem = ET.Element("annotation")
        ET.SubElement(ann_elem, "id").text = str(annotation.id)
        ET.SubElement(ann_elem, "type").text = str(annotation.type.name)

        for k, v in annotation.attributes.items():
            if k.isdigit():
                k = "_" + k
            ET.SubElement(ann_elem, k.replace(" ", "-")).text = str(v)

        ET.SubElement(ann_elem, "group").text = str(annotation.group)

        return ann_elem

    @staticmethod
    def _get_label(label_id: Optional[int], categories: Optional[CategoriesInfo]) -> str:
        label = ""
        if label_id is None:
            return ""
        if categories is not None:
            label_cat = categories.get(AnnotationType.label)
            if label_cat is not None:
                label = label_cat.items[label_id].name
        return label

    @classmethod
    def encode_label_object(
        cls, obj: Label, categories: Optional[CategoriesInfo]
    ) -> ET.ElementBase:
        ann_elem = cls.encode_annotation_base(obj)

        ET.SubElement(ann_elem, "label").text = str(cls._get_label(obj.label, categories))
        ET.SubElement(ann_elem, "label_id").text = str(obj.label)

        return ann_elem

    @classmethod
    def encode_mask_object(cls, obj: Mask, categories: Optional[CategoriesInfo]) -> ET.ElementBase:
        ann_elem = cls.encode_annotation_base(obj)

        ET.SubElement(ann_elem, "label").text = str(cls._get_label(obj.label, categories))
        ET.SubElement(ann_elem, "label_id").text = str(obj.label)

        return ann_elem

    @classmethod
    def encode_bbox_object(cls, obj: Bbox, categories: Optional[CategoriesInfo]) -> ET.ElementBase:
        ann_elem = cls.encode_annotation_base(obj)

        ET.SubElement(ann_elem, "label").text = str(cls._get_label(obj.label, categories))
        ET.SubElement(ann_elem, "label_id").text = str(obj.label)
        ET.SubElement(ann_elem, "x").text = str(obj.x)
        ET.SubElement(ann_elem, "y").text = str(obj.y)
        ET.SubElement(ann_elem, "w").text = str(obj.w)
        ET.SubElement(ann_elem, "h").text = str(obj.h)
        ET.SubElement(ann_elem, "area").text = str(obj.get_area())

        return ann_elem

    @classmethod
    def encode_points_object(
        cls, obj: Points, categories: Optional[CategoriesInfo]
    ) -> ET.ElementBase:
        ann_elem = cls.encode_annotation_base(obj)

        ET.SubElement(ann_elem, "label").text = str(cls._get_label(obj.label, categories))
        ET.SubElement(ann_elem, "label_id").text = str(obj.label)

        x, y, w, h = obj.get_bbox()
        area = w * h
        bbox_elem = ET.SubElement(ann_elem, "bbox")
        ET.SubElement(bbox_elem, "x").text = str(x)
        ET.SubElement(bbox_elem, "y").text = str(y)
        ET.SubElement(bbox_elem, "w").text = str(w)
        ET.SubElement(bbox_elem, "h").text = str(h)
        ET.SubElement(bbox_elem, "area").text = str(area)

        points = obj.points
        for i in range(0, len(points), 2):
            point_elem = ET.SubElement(ann_elem, "point")
            ET.SubElement(point_elem, "x").text = str(points[i])
            ET.SubElement(point_elem, "y").text = str(points[i + 1])
            ET.SubElement(point_elem, "visible").text = str(obj.visibility[i // 2].name)

        return ann_elem

    @classmethod
    def encode_polygon_object(
        cls, obj: Polygon, categories: Optional[CategoriesInfo]
    ) -> ET.ElementBase:
        ann_elem = cls.encode_annotation_base(obj)

        ET.SubElement(ann_elem, "label").text = str(cls._get_label(obj.label, categories))
        ET.SubElement(ann_elem, "label_id").text = str(obj.label)

        x, y, w, h = obj.get_bbox()
        area = w * h
        bbox_elem = ET.SubElement(ann_elem, "bbox")
        ET.SubElement(bbox_elem, "x").text = str(x)
        ET.SubElement(bbox_elem, "y").text = str(y)
        ET.SubElement(bbox_elem, "w").text = str(w)
        ET.SubElement(bbox_elem, "h").text = str(h)
        ET.SubElement(bbox_elem, "area").text = str(area)

        points = obj.points
        for i in range(0, len(points), 2):
            point_elem = ET.SubElement(ann_elem, "point")
            ET.SubElement(point_elem, "x").text = str(points[i])
            ET.SubElement(point_elem, "y").text = str(points[i + 1])

        return ann_elem

    @classmethod
    def encode_polyline_object(
        cls, obj: PolyLine, categories: Optional[CategoriesInfo]
    ) -> ET.ElementBase:
        ann_elem = cls.encode_annotation_base(obj)

        ET.SubElement(ann_elem, "label").text = str(cls._get_label(obj.label, categories))
        ET.SubElement(ann_elem, "label_id").text = str(obj.label)

        x, y, w, h = obj.get_bbox()
        area = w * h
        bbox_elem = ET.SubElement(ann_elem, "bbox")
        ET.SubElement(bbox_elem, "x").text = str(x)
        ET.SubElement(bbox_elem, "y").text = str(y)
        ET.SubElement(bbox_elem, "w").text = str(w)
        ET.SubElement(bbox_elem, "h").text = str(h)
        ET.SubElement(bbox_elem, "area").text = str(area)

        points = obj.points
        for i in range(0, len(points), 2):
            point_elem = ET.SubElement(ann_elem, "point")
            ET.SubElement(point_elem, "x").text = str(points[i])
            ET.SubElement(point_elem, "y").text = str(points[i + 1])

        return ann_elem

    @classmethod
    def encode_caption_object(cls, obj: Caption) -> ET.ElementBase:
        ann_elem = cls.encode_annotation_base(obj)

        ET.SubElement(ann_elem, "caption").text = str(obj.caption)

        return ann_elem

    @classmethod
    def encode_ellipse_object(
        cls, obj: Ellipse, categories: Optional[CategoriesInfo]
    ) -> ET.ElementBase:
        ann_elem = cls.encode_annotation_base(obj)

        ET.SubElement(ann_elem, "label").text = str(cls._get_label(obj.label, categories))
        ET.SubElement(ann_elem, "label_id").text = str(obj.label)

        ET.SubElement(ann_elem, "x1").text = str(obj.x1)
        ET.SubElement(ann_elem, "y1").text = str(obj.y1)
        ET.SubElement(ann_elem, "x2").text = str(obj.x2)
        ET.SubElement(ann_elem, "y2").text = str(obj.y2)
        ET.SubElement(ann_elem, "area").text = str(obj.get_area())

        return ann_elem

    @classmethod
    def encode_annotation(
        cls, o: Annotation, categories: Optional[CategoriesInfo] = None
    ) -> ET.ElementBase:
        if isinstance(o, Label):
            return cls.encode_label_object(o, categories)
        if isinstance(o, Mask):
            return cls.encode_mask_object(o, categories)
        if isinstance(o, Bbox):
            return cls.encode_bbox_object(o, categories)
        if isinstance(o, Points):
            return cls.encode_points_object(o, categories)
        if isinstance(o, PolyLine):
            return cls.encode_polyline_object(o, categories)
        if isinstance(o, Polygon):
            return cls.encode_polygon_object(o, categories)
        if isinstance(o, Caption):
            return cls.encode_caption_object(o)
        if isinstance(o, Ellipse):
            return cls.encode_ellipse_object(o, categories)
        if isinstance(o, HashKey):
            return cls.encode_annotation_base(o)

        raise NotImplementedError("Unexpected annotation object passed: %s" % o)

    @staticmethod
    def to_string(encoded_item: ET.ElementBase) -> str:
        return ET.tostring(encoded_item, encoding="unicode", pretty_print=True)


class XPathDatasetFilter(ItemTransform):
    def __init__(self, extractor: IDataset, xpath: str) -> None:
        super().__init__(extractor)

        try:
            xpath_eval = ET.XPath(xpath)
        except Exception:
            log.error("Failed to create XPath from expression '%s'", xpath)
            raise

        # Return true -> filter out an item
        self._f = lambda item: bool(
            xpath_eval(DatasetItemEncoder.encode(item, extractor.categories()))
        )

    def transform_item(self, item: DatasetItem) -> Optional[DatasetItem]:
        if not self._f(item):
            return None
        return item


class XPathAnnotationsFilter(ItemTransform):
    def __init__(self, extractor: IDataset, xpath: str, remove_empty: bool = False) -> None:
        super().__init__(extractor)

        try:
            xpath_eval = ET.XPath(xpath)
        except Exception:
            log.error("Failed to create XPath from expression '%s'", xpath)
            raise

        self._filter = xpath_eval

        self._remove_empty = remove_empty

    def transform_item(self, item: DatasetItem) -> Optional[DatasetItem]:
        if self._filter is None:
            return item

        encoded = DatasetItemEncoder.encode(item, self._extractor.categories())
        filtered = self._filter(encoded)
        filtered = [elem for elem in filtered if elem.tag == "annotation"]

        encoded = encoded.findall("annotation")
        annotations = [item.annotations[encoded.index(e)] for e in filtered]

        if self._remove_empty and len(annotations) == 0:
            return None
        return self.wrap_item(item, annotations=annotations)


class UserFunctionDatasetFilter(ItemTransform):
    """Filter dataset items using a user-provided Python function.

    Parameters:
        extractor: Datumaro `Dataset` to filter.
        filter_func: A Python callable that takes a `DatasetItem` as its input and
            returns a boolean. If the return value is True, that `DatasetItem` will be retained.
            Otherwise, it is removed.

    Example:
        This is an example of filtering dataset items with images larger than 1024 pixels::

        from datumaro.components.media import Image

        def filter_func(item: DatasetItem) -> bool:
            h, w = item.media_as(Image).size
            return h > 1024 or w > 1024

        filtered = UserFunctionDatasetFilter(
            extractor=dataset, filter_func=filter_func)
        # No items with an image height or width greater than 1024
        filtered_items = [item for item in filtered]
    """

    def __init__(self, extractor: IDataset, filter_func: Callable[[DatasetItem], bool]):
        super().__init__(extractor)

        self._filter_func = filter_func

    def transform_item(self, item: DatasetItem) -> Optional[DatasetItem]:
        return item if self._filter_func(item) else None


class UserFunctionAnnotationsFilter(ItemTransform):
    """Filter annotations using a user-provided Python function.

    Parameters:
        extractor: Datumaro `Dataset` to filter.
        filter_func: A Python callable that takes `DatasetItem` and `Annotation` as its inputs
            and returns a boolean. If the return value is True, the `Annotation` will be retained.
            Otherwise, it is removed.
        remove_empty: If True, `DatasetItem` without any annotations is removed
            after filtering its annotations. Otherwise, do not filter `DatasetItem`.

    Example:
        This is an example of removing bounding boxes sized greater than 50% of the image size::

        from datumaro.components.media import Image
        from datumaro.components.annotation import Annotation, Bbox

        def filter_func(item: DatasetItem, ann: Annotation) -> bool:
            # If the annotation is not a Bbox, do not filter
            if not isinstance(ann, Bbox):
                return False

            h, w = item.media_as(Image).size
            image_size = h * w
            bbox_size = ann.h * ann.w

            # Accept Bboxes smaller than 50% of the image size
            return bbox_size < 0.5 * image_size

        filtered = UserFunctionAnnotationsFilter(
            extractor=dataset, filter_func=filter_func)
        # No bounding boxes with a size greater than 50% of their image
        filtered_items = [item for item in filtered]
    """

    def __init__(
        self,
        extractor: IDataset,
        filter_func: Callable[[DatasetItem, Annotation], bool],
        remove_empty: bool = False,
    ):
        super().__init__(extractor)

        self._filter_func = filter_func
        self._remove_empty = remove_empty

    def transform_item(self, item: DatasetItem) -> Optional[DatasetItem]:
        filtered_anns = [ann for ann in item.annotations if self._filter_func(item, ann)]

        if self._remove_empty and not filtered_anns:
            return None
        return self.wrap_item(item, annotations=filtered_anns)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os

import numpy as np

from datumaro.components.annotation import HashKey
from datumaro.components.dataset import Dataset
from datumaro.components.media import MediaElement

templates = [
    "a photo of a {}.",
]

cifar10_templates = [
    "a photo of a {}.",
    "a blurry photo of a {}.",
    "a black and white photo of a {}.",
    "a low contrast photo of a {}.",
    "a high contrast photo of a {}.",
    "a bad photo of a {}.",
    "a good photo of a {}.",
    "a photo of a small {}.",
    "a photo of a big {}.",
    "a photo of the {}.",
    "a blurry photo of the {}.",
    "a black and white photo of the {}.",
    "a low contrast photo of the {}.",
    "a high contrast photo of the {}.",
    "a bad photo of the {}.",
    "a good photo of the {}.",
    "a photo of the small {}.",
    "a photo of the big {}.",
]

cifar100_templates = [
    "a photo of a {}.",
    "a blurry photo of a {}.",
    "a black and white photo of a {}.",
    "a low contrast photo of a {}.",
    "a high contrast photo of a {}.",
    "a bad photo of a {}.",
    "a good photo of a {}.",
    "a photo of a small {}.",
    "a photo of a big {}.",
    "a photo of the {}.",
    "a blurry photo of the {}.",
    "a black and white photo of the {}.",
    "a low contrast photo of the {}.",
    "a high contrast photo of the {}.",
    "a bad photo of the {}.",
    "a good photo of the {}.",
    "a photo of the small {}.",
    "a photo of the big {}.",
]

caltech101_templates = [
    "a photo of a {}.",
    "a painting of a {}.",
    "a plastic {}.",
    "a sculpture of a {}.",
    "a sketch of a {}.",
    "a tattoo of a {}.",
    "a toy {}.",
    "a rendition of a {}.",
    "a embroidered {}.",
    "a cartoon {}.",
    "a {} in a video game.",
    "a plushie {}.",
    "a origami {}.",
    "art of a {}.",
    "graffiti of a {}.",
    "a drawing of a {}.",
    "a doodle of a {}.",
    "a photo of the {}.",
    "a painting of the {}.",
    "the plastic {}.",
    "a sculpture of the {}.",
    "a sketch of the {}.",
    "a tattoo of the {}.",
    "the toy {}.",
    "a rendition of the {}.",
    "the embroidered {}.",
    "the cartoon {}.",
    "the {} in a video game.",
    "the plushie {}.",
    "the origami {}.",
    "art of the {}.",
    "graffiti of the {}.",
    "a drawing of the {}.",
    "a doodle of the {}.",
]

eurosat_templates = [
    "a centered satellite photo of {}.",
    "a centered satellite photo of a {}.",
    "a centered satellite photo of the {}.",
]

flowers101_templates = [
    "a photo of a {}, a type of flower.",
]

food101_templates = [
    "a photo of {}, a type of food.",
]

kitti_templates = [
    "{}",
]

kinetics_templates = [
    "a photo of {}.",
    "a photo of a person {}.",
    "a photo of a person using {}.",
    "a photo of a person doing {}.",
    "a photo of a person during {}.",
    "a photo of a person performing {}.",
    "a photo of a person practicing {}.",
    "a video of {}.",
    "a video of a person {}.",
    "a video of a person using {}.",
    "a video of a person doing {}.",
    "a video of a person during {}.",
    "a video of a person performing {}.",
    "a video of a person practicing {}.",
    "a example of {}.",
    "a example of a person {}.",
    "a example of a person using {}.",
    "a example of a person doing {}.",
    "a example of a person during {}.",
    "a example of a person performing {}.",
    "a example of a person practicing {}.",
    "a demonstration of {}.",
    "a demonstration of a person {}.",
    "a demonstration of a person using {}.",
    "a demonstration of a person doing {}.",
    "a demonstration of a person during {}.",
    "a demonstration of a person performing {}.",
    "a demonstration of a person practicing {}.",
]

mnist_templates = [
    'a photo of the number: "{}".',
]

format_templates = {
    "cifar10": cifar10_templates,
    "cifar100": cifar100_templates,
    "caltech101": caltech101_templates,
    "eurosat": eurosat_templates,
    "flowers101": flowers101_templates,
    "food101": food101_templates,
    "kitti": kitti_templates,
    "kinetics": kinetics_templates,
    "mnist": mnist_templates,
}


def select_uninferenced_dataset(dataset):
    uninferenced_dataset = Dataset(media_type=MediaElement, ann_types=set())
    for item in dataset:
        if not any(isinstance(annotation, HashKey) for annotation in item.annotations):
            uninferenced_dataset.put(item)
    return uninferenced_dataset


def calculate_hamming(B1, B2):
    """
    :param B1:  vector [n]
    :param B2:  vector [r*n]
    :return: hamming distance [r]
    """
    return np.count_nonzero(B1 != B2, axis=1)


def match_query_subset(query_id, dataset, subset=None):
    if subset:
        return dataset.get(query_id, subset)

    subset_names = dataset.subsets().keys()
    for subset_name in subset_names:
        try:
            query_datasetitem = dataset.get(query_id, subset_name)
            if query_datasetitem:
                return query_datasetitem
        except Exception:
            pass

    return None


def check_and_convert_to_list(paths):
    if isinstance(paths, str):
        paths = [paths]
    elif not isinstance(paths, list):
        raise ValueError("Invalid value type. Expected str or list.")

    valid_paths = []
    for path in paths:
        if os.path.exists(path):
            valid_paths.append(path)
        else:
            raise ValueError(f"Invalid path: {path}")

    return valid_paths



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from collections import defaultdict
from typing import Dict, List, Sequence, Tuple

from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.dataset_item_storage import DatasetItemStorage
from datumaro.components.merge import Merger

__all__ = ["UnionMerge"]


class UnionMerge(Merger):
    """
    Merge several datasets with "union" policy:

    - Label categories are merged according to the union of their label names.
    For example, if Dataset-A has {"car", "cat", "dog"} and Dataset-B has
    {"car", "bus", "truck"} labels, the merged dataset will have
    {"bust", "car", "cat", "dog", "truck"} labels.

    - If there are two or more dataset items whose (id, subset) pairs match each other,
    both are included in the merged dataset. At this time, since the same (id, subset)
    pair cannot be duplicated in the dataset, we add a suffix to the id of each source item.
    For example, if Dataset-A has DatasetItem(id="magic", subset="train") and Dataset-B has
    also DatasetItem(id="magic", subset="train"), the merged dataset will have
    DatasetItem(id="magic-0", subset="train") and DatasetItem(id="magic-1", subset="train").
    """

    def __init__(self, **options):
        super().__init__(**options)
        self._matching_table = {}

    def merge(self, sources: Sequence[IDataset]) -> DatasetItemStorage:
        dict_items: Dict[Tuple[str, str], List[DatasetItem]] = defaultdict(list)

        for source_idx, source in enumerate(sources):
            for item in source:
                if self._matching_table.get(source_idx, None):
                    for ann in item.annotations:
                        ann.label = self._matching_table[source_idx][ann.label]
                dict_items[item.id, item.subset].append(item)

        item_storage = DatasetItemStorage()

        for items in dict_items.values():
            if len(items) == 1:
                item_storage.put(items[0])
            else:
                for idx, item in enumerate(items):
                    # Add prefix
                    item_storage.put(item.wrap(id=f"{item.id}-{idx}"))

        return item_storage

    def merge_categories(self, sources: Sequence[IDataset]) -> Dict:
        dst_categories = {}

        label_cat = self._merge_label_categories(sources)
        if label_cat is None:
            label_cat = LabelCategories()
        dst_categories[AnnotationType.label] = label_cat

        return dst_categories

    def _merge_label_categories(self, sources: Sequence[IDataset]) -> LabelCategories:
        dst_cat = LabelCategories()
        for src_id, src_categories in enumerate(sources):
            src_cat = src_categories.get(AnnotationType.label)
            if src_cat is None:
                continue

            for src_label in src_cat.items:
                src_idx = src_cat.find(src_label.name)[0]
                dst_idx = dst_cat.find(src_label.name)[0]
                if dst_idx is None:
                    dst_cat.add(src_label.name, src_label.parent, src_label.attributes)
                    dst_idx = dst_cat.find(src_label.name)[0]

                if self._matching_table.get(src_id, None):
                    self._matching_table[src_id].update({src_idx: dst_idx})
                else:
                    self._matching_table[src_id] = {src_idx: dst_idx}

        return dst_cat



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List, Tuple

import numpy as np

from datumaro.components.abstracts import IModelInterpreter
from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred, PrepInfo
from datumaro.components.annotation import Annotation, Bbox, FeatureVector, Mask
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.media import Image


class SAMDecoderForBboxInterpreter(IModelInterpreter):
    h_model = 1024
    w_model = 1024
    onnx_mask_input = np.zeros((1, 1, 256, 256), dtype=np.float32)
    onnx_has_mask_input = np.zeros(1, dtype=np.float32)
    mask_threshold = 0.0

    def preprocess(self, inp: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        img = inp.media_as(Image).data

        img_embed = inp.annotations[-1]

        assert isinstance(
            img_embed, FeatureVector
        ), "annotations should have the image embedding vector as FeatureVector."

        h_img, w_img = img.shape[:2]

        scale = min(self.h_model / h_img, self.w_model / w_img)

        bboxes = [ann for ann in inp.annotations if isinstance(ann, Bbox)]
        onnx_coord = scale * np.array(
            [(bbox.x, bbox.y, bbox.x + bbox.w, bbox.y + bbox.h) for bbox in bboxes],
            dtype=np.float32,
        ).reshape(-1, 2, 2)
        onnx_label = np.array([2, 3] * len(onnx_coord)).reshape(-1, 2).astype(np.float32)

        decoder_inputs = {
            "image_embeddings": img_embed.vector[None, :, :, :],
            "point_coords": onnx_coord,
            "point_labels": onnx_label,
            "mask_input": self.onnx_mask_input,
            "has_mask_input": self.onnx_has_mask_input,
            "orig_im_size": np.array(img.shape[:2], dtype=np.float32),
        }

        return decoder_inputs, bboxes

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        """Postprocesses the outputs of the SAM decoder to generate masks associated with bounding boxes.

        Parameters:
            pred: List of dictionaries containing model predictions. Each dictionary should have the 'masks' key
                corresponding to the predicted mask of which shape is (1, H, W).
            info: List of associated bounding boxes obtained from the preprocessing step.

        Returns:
            List of `Mask`s associated with the input bounding boxes. Each Mask object contains the generated mask
            along with metadata from the corresponding `Bbox`:
            `id`, `group`, `object_id`, `label`, `z_order` and `attributes`.
        """
        masks = [np.squeeze(p["masks"], axis=0) > self.mask_threshold for p in pred]

        bboxes: List[Bbox] = info  # Return from preprocess()

        return [
            Mask(
                image=mask,
                id=bbox.id,
                group=bbox.group,
                object_id=bbox.object_id,
                label=bbox.label,
                z_order=bbox.z_order,
                attributes=bbox.attributes,
            )
            for bbox, mask in zip(bboxes, masks)
        ]

    def get_categories(self):
        return None



# Copyright (C) 2021-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

# The format is described here:
# https://docs.supervise.ly/data-organization/00_ann_format_navi

from __future__ import annotations

import logging as log
import os
import os.path as osp
import shutil
import uuid
from datetime import datetime

from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.dataset_item_storage import ItemStatus
from datumaro.components.errors import DatasetExportError, MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.media import PointCloud
from datumaro.util import cast, dump_json_file

from .format import PointCloudPath


class _SuperviselyPointCloudDumper:
    def __init__(self, extractor: IDataset, context: SuperviselyPointCloudExporter):
        self._extractor = extractor
        self._context = context

        timestamp = str(datetime.now())
        self._default_user_info = {
            "labelerLogin": "",
            "createdAt": timestamp,
            "updatedAt": timestamp,
        }

        self._key_id_data = {"tags": {}, "objects": {}, "figures": {}, "videos": {}}

        self._meta_data = {"classes": [], "tags": [], "projectType": "point_clouds"}

        # Meta info contents
        self._tag_meta = {}  # name -> descriptor

        # Registries of item annotations
        self._objects = {}  # id -> key

        self._label_cat = extractor.categories().get(AnnotationType.label, LabelCategories())

    def _write_related_images(self, item):
        img_dir = self._related_images_dir

        for idx, img in enumerate(item.media.extra_images):
            name = (
                osp.splitext(osp.basename(img.path))[0]
                if hasattr(img, "path")
                else f"extra_image_{idx}"
            )
            img_path = osp.join(img_dir, item.id + "_pcd", name + self._find_image_ext(img))
            if img.has_data:
                img.save(img_path)

            img_data = {
                "name": osp.basename(img_path),
                "meta": {
                    "sensorsData": {
                        "extrinsicMatrix": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                        "intrinsicMatrix": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    }
                },
            }

            dump_json_file(osp.join(img_dir, img_path + ".json"), img_data, indent=True)

    def _write_pcd(self, item):
        self._context._save_point_cloud(item, basedir=self._point_cloud_dir)

    def _write_meta(self):
        for tag in self._tag_meta.values():
            if tag["value_type"] is None:
                tag["value_type"] = "any_string"
            tag["classes"] = list(tag["classes"])
        self._meta_data["tags"] = list(self._tag_meta.values())

        dump_json_file(
            osp.join(self._save_dir, PointCloudPath.META_FILE), self._meta_data, indent=True
        )

    def _write_key_id(self):
        objects = self._objects
        key_id_data = self._key_id_data

        key_id_data["objects"] = {v: k for k, v in objects.items()}

        dump_json_file(
            osp.join(self._save_dir, PointCloudPath.KEY_ID_FILE), key_id_data, indent=True
        )

    def _write_item_annotations(self, item):
        key_id_data = self._key_id_data

        item_id = cast(item.attributes.get("frame"), int)
        if item_id is None or self._context._reindex:
            item_id = len(key_id_data["videos"]) + 1

        item_key = str(uuid.uuid4())
        key_id_data["videos"][item_key] = item_id

        item_user_info = {
            k: item.attributes.get(k, default_v) for k, default_v in self._default_user_info.items()
        }

        item_ann_data = {
            "description": item.attributes.get("description", ""),
            "key": item_key,
            "tags": [],
            "objects": [],
            "figures": [],
        }
        self._export_item_attributes(item, item_ann_data, item_user_info)
        self._export_item_annotations(item, item_ann_data, item_user_info)

        ann_path = osp.join(self._ann_dir, item.id + ".pcd.json")
        os.makedirs(osp.dirname(ann_path), exist_ok=True)
        dump_json_file(ann_path, item_ann_data, indent=True)

    def _export_item_attributes(self, item, item_ann_data, item_user_info):
        for attr_name, attr_value in item.attributes.items():
            if attr_name in PointCloudPath.SPECIAL_ATTRS:
                continue

            attr_value = self._encode_attr_value(attr_value)

            tag = self._register_tag(attr_name, value=attr_value, applicable_type="imagesOnly")

            if tag["applicable_type"] != "imagesOnly":
                tag["applicable_type"] = "all"

            value_type = self._define_attr_type(attr_value)
            if tag["value_type"] is None:
                tag["value_type"] = value_type
            elif tag["value_type"] != value_type:
                raise DatasetExportError(
                    "Item %s: mismatching "
                    "value types for tag %s: %s vs %s"
                    % (item.id, attr_name, tag["value_type"], value_type)
                )

            tag_key = str(uuid.uuid4())
            item_ann_data["tags"].append(
                {
                    "key": tag_key,
                    "name": attr_name,
                    "value": attr_value,
                    **item_user_info,
                }
            )

            # only item attributes are listed in the key_id file
            # meta tag ids have no relation to key_id tag ids!
            tag_id = len(self._key_id_data["tags"]) + 1
            self._key_id_data["tags"][tag_key] = tag_id

    def _export_item_annotations(self, item, item_ann_data, item_user_info):
        objects = self._objects
        tags = self._tag_meta
        label_cat = self._label_cat
        key_id_data = self._key_id_data

        image_objects = set()
        for ann in item.annotations:
            if not ann.type == AnnotationType.cuboid_3d:
                continue

            obj_id = cast(ann.attributes.get("track_id", ann.id), int)
            if obj_id is None:
                # should not be affected by reindex
                # because it is used to match figures,
                # including different frames
                obj_id = len(self._objects) + 1

            object_key = objects.setdefault(obj_id, str(uuid.uuid4()))
            object_label = label_cat[ann.label].name
            if obj_id not in image_objects:
                ann_user_info = {
                    k: ann.attributes.get(k, default_v) for k, default_v in item_user_info.items()
                }

                obj_ann_data = {
                    "key": object_key,
                    "classTitle": object_label,
                    "tags": [],
                    "objects": [],
                    "figures": [],
                    **ann_user_info,
                }

                for attr_name, attr_value in ann.attributes.items():
                    if attr_name in PointCloudPath.SPECIAL_ATTRS:
                        continue

                    attr_value = self._encode_attr_value(attr_value)

                    tag = tags.get(attr_name)
                    if tag is None:
                        if self._context._allow_undeclared_attrs:
                            tag = self._register_tag(attr_name, applicable_type="objectsOnly")
                            tags[attr_name] = tag
                        else:
                            log.warning(
                                "Item %s: skipping undeclared "
                                "attribute '%s' for label '%s' "
                                "(allow with --allow-undeclared-attrs option)",
                                item.id,
                                attr_name,
                                object_label,
                            )
                            continue

                    if tag["applicable_type"] == "imagesOnly":
                        tag["applicable_type"] = "all"
                    elif tag["applicable_type"] == "objectsOnly" and tag["classes"]:
                        tag["classes"].add(object_label)

                    value_type = self._define_attr_type(attr_value)
                    if tag["value_type"] is None:
                        tag["value_type"] = value_type
                    elif tag["value_type"] != value_type:
                        raise DatasetExportError(
                            "Item %s: mismatching "
                            "value types for tag %s: %s vs %s"
                            % (item.id, attr_name, tag["value_type"], value_type)
                        )

                    tag_key = str(uuid.uuid4())
                    obj_ann_data["tags"].append(
                        {
                            "key": tag_key,
                            "name": attr_name,
                            "value": attr_value,
                            **ann_user_info,
                        }
                    )

                item_ann_data["objects"].append(obj_ann_data)

                image_objects.add(obj_id)

            figure_key = str(uuid.uuid4())
            item_ann_data["figures"].append(
                {
                    "key": figure_key,
                    "objectKey": object_key,
                    "geometryType": "cuboid_3d",
                    "geometry": {
                        "position": {
                            "x": float(ann.position[0]),
                            "y": float(ann.position[1]),
                            "z": float(ann.position[2]),
                        },
                        "rotation": {
                            "x": float(ann.rotation[0]),
                            "y": float(ann.rotation[1]),
                            "z": float(ann.rotation[2]),
                        },
                        "dimensions": {
                            "x": float(ann.scale[0]),
                            "y": float(ann.scale[1]),
                            "z": float(ann.scale[2]),
                        },
                    },
                    **ann_user_info,
                }
            )
            figure_id = ann.id
            if self._context._reindex or figure_id is None:
                figure_id = len(key_id_data["figures"]) + 1
            key_id_data["figures"][figure_key] = figure_id

    @staticmethod
    def _encode_attr_value(v):
        if v is True or v is False:  # use is to check the type too
            v = str(v).lower()
        return v

    @staticmethod
    def _define_attr_type(v):
        if isinstance(v, (int, float)):
            t = "any_number"
        else:
            t = "any_string"
        return t

    def _register_tag(self, name, **kwargs):
        tag = {
            "name": name,
            "value_type": None,
            "color": "",
            "id": len(self._tag_meta) + 1,
            "hotkey": "",
            "applicable_type": "all",
            "classes": set(),
        }
        tag.update(kwargs)
        return self._tag_meta.setdefault(name, tag)

    def _make_dirs(self):
        save_dir = self._context._save_dir
        os.makedirs(save_dir, exist_ok=True)
        self._save_dir = save_dir

        base_dir = osp.join(self._save_dir, PointCloudPath.BASE_DIR)
        os.makedirs(base_dir, exist_ok=True)

        ann_dir = osp.join(base_dir, PointCloudPath.ANNNOTATION_DIR)
        os.makedirs(ann_dir, exist_ok=True)
        self._ann_dir = ann_dir

        point_cloud_dir = osp.join(base_dir, PointCloudPath.POINT_CLOUD_DIR)
        os.makedirs(point_cloud_dir, exist_ok=True)
        self._point_cloud_dir = point_cloud_dir

        related_images_dir = osp.join(base_dir, PointCloudPath.RELATED_IMAGES_DIR)
        os.makedirs(related_images_dir, exist_ok=True)
        self._related_images_dir = related_images_dir

    def _init_meta(self):
        for attr in self._label_cat.attributes:
            self._register_tag(attr, applicable_type="objectsOnly")

        for idx, label in enumerate(self._label_cat):
            self._meta_data["classes"].append(
                {
                    "id": idx,
                    "title": label.name,
                    "color": "",
                    "shape": "cuboid_3d",
                    "geometry_config": {},
                }
            )

            for attr in label.attributes:
                tag = self._register_tag(attr, applicable_type="objectsOnly")
                tag["classes"].add(label.name)

    def _find_image_ext(self, image):
        src_ext = image.ext
        return self._context._image_ext or src_ext or self._context._default_image_ext

    def dump(self):
        self._make_dirs()

        self._init_meta()

        for item in self._context._extractor:
            if self._context._save_media:
                if item.media:
                    self._write_pcd(item)
                else:
                    log.debug("Item '%s' has no point cloud info", item.id)

                if item.media and item.media.extra_images:
                    self._write_related_images(item)
                else:
                    log.debug("Item '%s' has no related images info", item.id)

            self._write_item_annotations(item)

        self._write_meta()
        self._write_key_id()


class SuperviselyPointCloudExporter(Exporter):
    NAME = "sly_pointcloud"
    DEFAULT_IMAGE_EXT = PointCloudPath.DEFAULT_IMAGE_EXT

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--reindex",
            action="store_true",
            help="Assign new indices to frames (default: %(default)s)",
        )
        parser.add_argument(
            "--allow-undeclared-attrs",
            action="store_true",
            help="Write annotation attributes even if they are not present in "
            "the input dataset metainfo (default: %(default)s)",
        )
        return parser

    def __init__(self, extractor, save_dir, reindex=False, allow_undeclared_attrs=False, **kwargs):
        super().__init__(extractor, save_dir, **kwargs)

        self._reindex = reindex
        self._allow_undeclared_attrs = allow_undeclared_attrs

    def _apply_impl(self):
        if self._extractor.media_type() and self._extractor.media_type() is not PointCloud:
            raise MediaTypeError("Media type is not an image")

        if 1 < len(self._extractor.subsets()):
            log.warning(
                "Supervisely pointcloud format supports only a single "
                "subset. Subset information will be ignored on export."
            )

        _SuperviselyPointCloudDumper(self._extractor, self).dump()

    @classmethod
    def patch(cls, dataset, patch, save_dir, **kwargs):
        conv = cls(patch.as_dataset(dataset), save_dir=save_dir, **kwargs)
        conv.apply()

        for (item_id, subset), status in patch.updated_items.items():
            if status != ItemStatus.removed:
                item = patch.data.get(item_id, subset)
            else:
                item = DatasetItem(item_id, subset=subset)

            if not (status == ItemStatus.removed or not item.media):
                continue

            pcd_name = conv._make_pcd_filename(item)

            ann_path = osp.join(
                save_dir,
                PointCloudPath.BASE_DIR,
                PointCloudPath.ANNNOTATION_DIR,
                pcd_name + ".json",
            )
            if osp.isfile(ann_path):
                os.remove(ann_path)

            pcd_path = osp.join(
                save_dir, PointCloudPath.BASE_DIR, PointCloudPath.POINT_CLOUD_DIR, pcd_name
            )
            if osp.isfile(pcd_path):
                os.remove(pcd_path)

            images_dir = osp.join(
                save_dir,
                PointCloudPath.BASE_DIR,
                PointCloudPath.RELATED_IMAGES_DIR,
                osp.splitext(pcd_name)[0] + "_pcd",
            )
            if osp.isdir(images_dir):
                shutil.rmtree(images_dir)



# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

# Implements MOTS format https://www.vision.rwth-aachen.de/page/mots

import logging as log
import os
import os.path as osp
from enum import Enum
from glob import iglob
from typing import List, Optional

import numpy as np

from datumaro.components.annotation import AnnotationType, ExtractedMask, LabelCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.image import find_images, lazy_image, save_image
from datumaro.util.mask_tools import merge_masks
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file


class MotsPath:
    MASKS_DIR = "instances"
    IMAGE_DIR = "images"
    IMAGE_EXT = ".jpg"
    LABELS_FILE = "labels.txt"
    MAX_INSTANCES = 1000


class MotsLabels(Enum):
    background = 0
    car = 1
    pedestrian = 2
    ignored = 10


class MotsPngExtractor(SubsetBase):
    @staticmethod
    def detect_dataset(path):
        if osp.isdir(osp.join(path, MotsPath.MASKS_DIR)):
            return [{"url": path, "format": MotsPngExtractor.NAME}]
        return []

    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        assert osp.isdir(path), path
        super().__init__(subset=subset, ctx=ctx)
        self._images_dir = osp.join(path, "images")
        self._anno_dir = osp.join(path, MotsPath.MASKS_DIR)
        if has_meta_file(path):
            self._categories = {
                AnnotationType.label: LabelCategories.from_iterable(parse_meta_file(path).keys())
            }
        else:
            self._categories = self._parse_categories(
                osp.join(self._anno_dir, MotsPath.LABELS_FILE)
            )
        self._items = self._parse_items()

    def _parse_categories(self, path):
        if osp.isfile(path):
            labels = []
            with open(path, encoding="utf-8") as f:
                for label in f:
                    label = label.strip()
                    if label:
                        labels.append(label)
        else:
            labels = [l.name for l in MotsLabels]
        return {AnnotationType.label: LabelCategories.from_iterable(labels)}

    def _parse_items(self):
        items = []

        image_dir = self._images_dir
        if osp.isdir(image_dir):
            images = {
                osp.splitext(osp.relpath(p, image_dir))[0]: p
                for p in find_images(image_dir, recursive=True)
            }
        else:
            images = {}

        for p in sorted(iglob(self._anno_dir + "/**/*.png", recursive=True)):
            item_id = osp.splitext(osp.relpath(p, self._anno_dir))[0]
            image = images.get(item_id)
            if image:
                image = Image.from_file(path=image)
            items.append(
                DatasetItem(
                    id=item_id,
                    subset=self._subset,
                    media=image,
                    annotations=self._parse_annotations(p),
                )
            )
        return items

    @staticmethod
    def _lazy_extract_mask(mask, v):
        return lambda: mask == v

    def _parse_annotations(self, path):
        index_mask = lazy_image(path, dtype=np.uint16)
        np_index_mask = index_mask()

        masks = []
        for obj_id in np.unique(np_index_mask):
            class_id, instance_id = divmod(obj_id, MotsPath.MAX_INSTANCES)
            z_order = 0
            if class_id == 0:
                continue  # background
            if class_id == 10 and len(self._categories[AnnotationType.label]) < 10:
                z_order = 1
                class_id = self._categories[AnnotationType.label].find(MotsLabels.ignored.name)[0]
            else:
                class_id -= 1
            masks.append(
                ExtractedMask(
                    index_mask=index_mask,
                    index=obj_id,
                    label=class_id,
                    z_order=z_order,
                    attributes={"track_id": instance_id},
                )
            )
            self._ann_types.add(AnnotationType.mask)
        return masks


class MotsImporter(Importer):
    @classmethod
    def find_sources(cls, path):
        if not osp.isdir(path):
            return []

        subsets = MotsPngExtractor.detect_dataset(path)
        if not subsets:
            for p in os.listdir(path):
                detected = MotsPngExtractor.detect_dataset(osp.join(path, p))
                for s in detected:
                    s.setdefault("options", {})["subset"] = p
                subsets.extend(detected)
        return subsets

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [".png", ".txt"]


class MotsPngExporter(Exporter):
    DEFAULT_IMAGE_EXT = MotsPath.IMAGE_EXT

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        os.makedirs(self._save_dir, exist_ok=True)

        if self._save_dataset_meta:
            self._save_meta_file(self._save_dir)

        for subset_name, subset in self._extractor.subsets().items():
            subset_dir = osp.join(self._save_dir, subset_name)
            image_dir = osp.join(subset_dir, MotsPath.IMAGE_DIR)
            anno_dir = osp.join(subset_dir, MotsPath.MASKS_DIR)
            os.makedirs(anno_dir, exist_ok=True)

            for item in subset:
                log.debug("Converting item '%s'", item.id)

                if self._save_media:
                    if item.media and item.media.has_data:
                        self._save_image(item, subdir=image_dir)
                    else:
                        log.debug("Item '%s' has no image", item.id)

                self._save_annotations(item, anno_dir)

            with open(osp.join(anno_dir, MotsPath.LABELS_FILE), "w", encoding="utf-8") as f:
                f.write("\n".join(l.name for l in subset.categories()[AnnotationType.label].items))

    def _save_annotations(self, item, anno_dir):
        masks = [a for a in item.annotations if a.type == AnnotationType.mask]
        if not masks:
            return

        instance_ids = [int(a.attributes["track_id"]) for a in masks]
        masks = sorted(zip(masks, instance_ids), key=lambda e: e[0].z_order)
        mask = merge_masks(
            (m.image, MotsPath.MAX_INSTANCES * (1 + m.label) + id) for m, id in masks
        )
        save_image(osp.join(anno_dir, item.id + ".png"), mask, create_dir=True, dtype=np.uint16)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import re
from typing import Union


def to_bytes(size: Union[str, int]) -> int:
    if isinstance(size, int):
        return size
    assert isinstance(size, str)

    match = re.match(r"^([\d\.]+)\s*([a-zA-Z]{0,3})$", size.strip())

    if match is None:
        raise ValueError(f"Cannot parse {size} string.")

    units = {
        "": 1,
        "B": 1,
        "KIB": 2**10,
        "MIB": 2**20,
        "GIB": 2**30,
        "TIB": 2**40,
        "KB": 10**3,
        "MB": 10**6,
        "GB": 10**9,
        "TB": 10**12,
        "K": 2**10,
        "M": 2**20,
        "G": 2**30,
        "T": 2**40,
    }

    number, unit = int(match.group(1)), match.group(2).upper()

    if unit not in units:
        raise ValueError(f"{size} has disallowed unit ({unit}).")

    return number * units[unit]



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT


import importlib.util
import sys
from types import ModuleType


def lazy_import(name: str) -> ModuleType:
    spec = importlib.util.find_spec(name)
    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    return module



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from collections import OrderedDict
from copy import deepcopy
from typing import List, Optional

import numpy as np
from defusedxml import ElementTree

from datumaro.components.annotation import (
    AnnotationType,
    Bbox,
    Label,
    LabelCategories,
    Mask,
    Points,
    Polygon,
    PolyLine,
)
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import DatasetImportError
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util import mask_tools

from .format import CvatPath


def _find_meta_root(path: str):
    context = ElementTree.iterparse(path, events=("start", "end"))
    context = iter(context)

    meta_root = None

    for event, elem in context:
        if elem.tag == "meta" and event == "start":
            meta_root = elem
        elif elem.tag == "meta" and event == "end":
            break

    if meta_root is None:
        raise DatasetImportError("CVAT XML file should have <meta> tag.")

    return meta_root, context


class CvatBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        assert osp.isfile(path), path
        rootpath = osp.dirname(path)
        images_dir = ""
        if osp.isdir(osp.join(rootpath, CvatPath.IMAGES_DIR)):
            images_dir = osp.join(rootpath, CvatPath.IMAGES_DIR)
        self._images_dir = images_dir
        self._path = path

        if not subset:
            subset = osp.splitext(osp.basename(path))[0]
        super().__init__(subset=subset, ctx=ctx)

        items, categories = self._parse(path)
        self._categories = categories
        self._items = list(self._load_items(items).values())

    def _parse(self, path):
        meta_root, context = _find_meta_root(path)

        categories, frame_size, attribute_types = self._parse_meta(meta_root)

        items = OrderedDict()

        track = None
        shape = None
        tag = None
        attributes = None
        image = None
        subset = None
        for ev, el in context:
            if ev == "start":
                if el.tag == "track":
                    track = {
                        "id": el.attrib["id"],
                        "label": el.attrib.get("label"),
                        "group": int(el.attrib.get("group_id", 0)),
                        "height": frame_size[0],
                        "width": frame_size[1],
                    }
                    subset = el.attrib.get("subset")
                elif el.tag == "image":
                    image = {
                        "name": el.attrib.get("name"),
                        "frame": el.attrib["id"],
                        "width": el.attrib.get("width"),
                        "height": el.attrib.get("height"),
                    }
                    subset = el.attrib.get("subset")
                elif el.tag in CvatPath.SUPPORTED_IMPORT_SHAPES and (track or image):
                    attributes = {}
                    shape = {
                        "type": None,
                        "attributes": attributes,
                    }
                    if track:
                        shape.update(track)
                        shape["track_id"] = int(track["id"])
                    if image:
                        shape.update(image)
                elif el.tag == "tag" and image:
                    attributes = {}
                    tag = {
                        "frame": image["frame"],
                        "attributes": attributes,
                        "group": int(el.attrib.get("group_id", 0)),
                        "label": el.attrib["label"],
                    }
            elif ev == "end":
                if el.tag == "attribute" and attributes is not None:
                    attr_value = el.text or ""
                    attr_type = attribute_types.get(el.attrib["name"])
                    if el.text in ["true", "false"]:
                        attr_value = attr_value == "true"
                    elif attr_type is not None and attr_type != "text":
                        try:
                            attr_value = float(attr_value)
                        except ValueError:
                            pass
                    attributes[el.attrib["name"]] = attr_value
                elif el.tag in CvatPath.SUPPORTED_IMPORT_SHAPES:
                    if track is not None:
                        shape["frame"] = el.attrib["frame"]
                        shape["outside"] = el.attrib.get("outside") == "1"
                        shape["keyframe"] = el.attrib.get("keyframe") == "1"
                    if image is not None:
                        shape["label"] = el.attrib.get("label")
                        shape["group"] = int(el.attrib.get("group_id", 0))

                    shape["type"] = el.tag
                    shape["occluded"] = el.attrib.get("occluded") == "1"
                    shape["z_order"] = int(el.attrib.get("z_order", 0))

                    if el.tag == "box":
                        shape["points"] = list(
                            map(
                                float,
                                [
                                    el.attrib["xtl"],
                                    el.attrib["ytl"],
                                    el.attrib["xbr"],
                                    el.attrib["ybr"],
                                ],
                            )
                        )
                    elif el.tag == "mask":
                        shape["rle"] = el.attrib["rle"]
                        shape["left"] = el.attrib["left"]
                        shape["top"] = el.attrib["top"]
                        shape["width"] = el.attrib["width"]
                        shape["height"] = el.attrib["height"]
                    else:
                        shape["points"] = []
                        for pair in el.attrib["points"].split(";"):
                            shape["points"].extend(map(float, pair.split(",")))

                    if subset is None or subset == self._subset:
                        frame_desc = items.get(shape["frame"], {"annotations": []})
                        frame_desc["annotations"].append(
                            self._parse_shape_ann(shape, categories, image)
                        )
                        items[shape["frame"]] = frame_desc

                    shape = None

                elif el.tag == "tag":
                    if subset is None or subset == self._subset:
                        frame_desc = items.get(tag["frame"], {"annotations": []})
                        frame_desc["annotations"].append(self._parse_tag_ann(tag, categories))
                        items[tag["frame"]] = frame_desc
                    tag = None
                elif el.tag == "track":
                    track = None
                elif el.tag == "image":
                    if subset is None or subset == self._subset:
                        frame_desc = items.get(image["frame"], {"annotations": []})
                        frame_desc.update(
                            {
                                "name": image.get("name"),
                                "height": image.get("height"),
                                "width": image.get("width"),
                            }
                        )
                        items[image["frame"]] = frame_desc
                    image = None
                el.clear()

        return items, categories

    @staticmethod
    def _parse_meta(meta_root):
        categories = {}

        frame_size = None
        original_size = [item for item in meta_root.iter("original_size")]

        if len(original_size) > 1:
            raise DatasetImportError("CVAT XML file should have only one <original_size> tag.")
        elif len(original_size) == 1:
            frame_size = (
                int(original_size[0].find("height").text),
                int(original_size[0].find("width").text),
            )

        mode = None
        labels = OrderedDict()

        for label in meta_root.iter("label"):
            name = label.find("name").text
            labels[name] = [
                {
                    "name": attr.find("name").text,
                    "input_type": attr.find("input_type").text,
                }
                for attr in label.iter("attribute")
            ]

        common_attrs = ["occluded"]
        if mode == "interpolation":
            common_attrs.append("keyframe")
            common_attrs.append("outside")
            common_attrs.append("track_id")

        label_cat = LabelCategories(attributes=common_attrs)
        attribute_types = {}
        for label, attrs in labels.items():
            attr_names = {v["name"] for v in attrs}
            label_cat.add(label, attributes=attr_names)
            for attr in attrs:
                attribute_types[attr["name"]] = attr["input_type"]

        categories[AnnotationType.label] = label_cat
        return categories, frame_size, attribute_types

    @classmethod
    def _parse_shape_ann(cls, ann, categories, image):
        ann_id = ann.get("id", 0)
        ann_type = ann["type"]

        attributes = ann.get("attributes") or {}
        if "occluded" in categories[AnnotationType.label].attributes:
            attributes["occluded"] = ann.get("occluded", False)
        if "outside" in ann:
            attributes["outside"] = ann["outside"]
        if "keyframe" in ann:
            attributes["keyframe"] = ann["keyframe"]
        if "track_id" in ann:
            attributes["track_id"] = ann["track_id"]

        group = ann.get("group")

        label = ann.get("label")
        label_id = categories[AnnotationType.label].find(label)[0]

        z_order = ann.get("z_order", 0)
        points = ann.get("points", [])

        if ann_type == "polyline":
            return PolyLine(
                points,
                label=label_id,
                z_order=z_order,
                id=ann_id,
                attributes=attributes,
                group=group,
            )

        elif ann_type == "polygon":
            return Polygon(
                points,
                label=label_id,
                z_order=z_order,
                id=ann_id,
                attributes=attributes,
                group=group,
            )

        elif ann_type == "points":
            return Points(
                points,
                label=label_id,
                z_order=z_order,
                id=ann_id,
                attributes=attributes,
                group=group,
            )

        elif ann_type == "box":
            x, y = points[0], points[1]
            w, h = points[2] - x, points[3] - y
            return Bbox(
                x,
                y,
                w,
                h,
                label=label_id,
                z_order=z_order,
                id=ann_id,
                attributes=attributes,
                group=group,
            )

        elif ann_type == "mask":
            rle = ann.get("rle")
            mask_w, mask_h = int(ann.get("width")), int(ann.get("height"))
            mask_l, mask_t = int(ann.get("left")), int(ann.get("top"))
            img_w, img_h = int(image.get("width")), int(image.get("height"))

            rle_uncompressed = {
                "counts": np.array([int(str_num) for str_num in rle.split(",")], dtype=np.uint32),
                "size": np.array([mask_w, mask_h]),
            }

            def _gen_mask():
                # From the manual test for the dataset exported from the CVAT 2.5,
                # the RLE encoding in the dataset has (W, H) binary 2D np.ndarray, not (H, W)
                # Therefore, we need to tranpose it to make its shape as (H, W).
                mask = mask_tools.rle_to_mask(rle_uncompressed).transpose()
                canvas = np.zeros(shape=[img_h, img_w], dtype=np.uint8)
                canvas[mask_t : mask_t + mask_h, mask_l : mask_l + mask_w] = mask
                return canvas

            return Mask(
                image=_gen_mask,
                label=label_id,
                z_order=z_order,
                id=ann_id,
                attributes=attributes,
                group=group,
            )

        else:
            raise NotImplementedError("Unknown annotation type '%s'" % ann_type)

    @classmethod
    def _parse_tag_ann(cls, ann, categories):
        label = ann.get("label")
        label_id = categories[AnnotationType.label].find(label)[0]
        group = ann.get("group")
        attributes = ann.get("attributes")
        return Label(label_id, attributes=attributes, group=group)

    def _load_items(self, parsed):
        for frame_id, item_desc in parsed.items():
            name = item_desc.get("name", "frame_%06d.png" % int(frame_id))

            image_path_opt_1 = osp.join(self._images_dir, name)
            image_path_opt_2 = (
                osp.join(self._images_dir, self._subset, name) if self._subset is not None else None
            )
            if osp.exists(image_path_opt_1):
                image = image_path_opt_1
            elif image_path_opt_2 and osp.exists(image_path_opt_2):
                image = image_path_opt_2
            elif "name" not in item_desc:
                # If --use-track flag is on
                # TODO: Revisit all the CVAT import/export parts.
                image = image_path_opt_1
            else:
                raise DatasetImportError(f"Cannot find an image which has name={name}.")

            image_size = (item_desc.get("height"), item_desc.get("width"))
            if all(image_size):
                image = Image.from_file(path=image, size=tuple(map(int, image_size)))
            else:
                image = Image.from_file(path=image)

            parsed[frame_id] = DatasetItem(
                id=osp.splitext(name)[0],
                subset=self._subset,
                media=image,
                annotations=item_desc.get("annotations"),
                attributes={"frame": int(frame_id)},
            )
            for ann in item_desc.get("annotations"):
                self._ann_types.add(ann.type)

        return parsed


class CvatImporter(Importer):
    _ANNO_EXT = ".xml"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        annot_file = context.require_file(f"*{cls._ANNO_EXT}")

        with context.probe_text_file(
            annot_file,
            'must be an XML file with an "annotations" root element',
        ) as f:
            _, root_elem = next(ElementTree.iterparse(f, events=("start",)))
            if root_elem.tag != "annotations":
                raise Exception

    @staticmethod
    def find_subsets(meta_root):
        subsets = [item.text for item in meta_root.iter("subset")]
        if len(subsets) == 0:
            raise DatasetImportError("CVAT XML should include <subset> tags.")
        return subsets

    @classmethod
    def find_sources(cls, path):
        source_files = cls._find_sources_recursive(path, ".xml", "cvat")
        sources = []

        for source in source_files:
            path = source["url"]
            meta_root, _ = _find_meta_root(path)

            if meta_root.find("project") is not None:
                for subset in cls.find_subsets(meta_root):
                    source_clone = deepcopy(source)
                    source_clone["options"] = {"subset": subset}
                    sources += [source_clone]
            elif meta_root.find("task") is not None:
                sources += [source]
            else:
                raise DatasetImportError(
                    "CVAT XML file should have a <meta> -> <task> or <meta> -> <project> subtree."
                )

        return sources

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import re

emoji_pattern = re.compile(
    "|".join(
        [
            "[\U0001F600-\U0001F64F]",  # emoticons
            "[\U0001F300-\U0001F5FF]",  # symbols & pictographs
            "[\U0001F680-\U0001F6FF]",  # transport & map symbols
            "[\U0001F1E0-\U0001F1FF]",  # flags (iOS)
            "[\u2600-\u26FF]",  # Miscellaneous Symbols
            "[\u2700-\u27BF]",  # Dingbats
            "[\U0001F900-\U0001F9FF]",  # Supplemental Symbols and Pictographs
        ]
    ),
    flags=re.UNICODE,
)



# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import os
import os.path as osp
import re
from typing import List, Optional

from datumaro.components.annotation import AnnotationType, Bbox, Label, LabelCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util import str_to_bool
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file


class WiderFacePath:
    IMAGE_EXT = ".jpg"
    ANNOTATIONS_DIR = "wider_face_split"
    IMAGES_DIR = "images"
    SUBSET_DIR = "WIDER_"
    LABELS_FILE = "labels.txt"
    IMAGES_DIR_NO_LABEL = "no_label"
    BBOX_ATTRIBUTES = ["blur", "expression", "illumination", "occluded", "pose", "invalid"]
    DEFAULT_LABEL = "face"


class WiderFaceBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isfile(path):
            raise FileNotFoundError(errno.ENOENT, "Can't find annotations file", path)
        self._path = path
        self._dataset_dir = osp.dirname(osp.dirname(path))

        if not subset:
            subset = osp.splitext(osp.basename(path))[0]
            if re.fullmatch(r"wider_face_\S+((_bbx_gt)|(_filelist))", subset):
                subset = subset.split("_")[2]

        super().__init__(subset=subset, ctx=ctx)

        self._categories = self._load_categories()
        self._items = list(self._load_items(path).values())

    def _load_categories(self):
        label_cat = LabelCategories()
        if has_meta_file(self._dataset_dir):
            labels = parse_meta_file(self._dataset_dir).keys()
            for label in labels:
                label_cat.add(label)
        elif osp.isfile(osp.join(self._dataset_dir, WiderFacePath.LABELS_FILE)):
            path = osp.join(self._dataset_dir, WiderFacePath.LABELS_FILE)
            with open(path, encoding="utf-8") as labels_file:
                for line in labels_file:
                    label_cat.add(line.strip())
        else:
            label_cat.add(WiderFacePath.DEFAULT_LABEL)
            subset_path = osp.join(
                self._dataset_dir, WiderFacePath.SUBSET_DIR + self._subset, WiderFacePath.IMAGES_DIR
            )
            if osp.isdir(subset_path):
                for images_dir in sorted(os.listdir(subset_path)):
                    if (
                        osp.isdir(osp.join(subset_path, images_dir))
                        and images_dir != WiderFacePath.IMAGES_DIR_NO_LABEL
                    ):
                        if "--" in images_dir:
                            images_dir = images_dir.split("--")[1]
                        if images_dir != WiderFacePath.DEFAULT_LABEL:
                            label_cat.add(images_dir)
            if len(label_cat) == 1:
                label_cat = LabelCategories()
        return {AnnotationType.label: label_cat}

    def _load_items(self, path):
        items = {}
        label_categories = self._categories[AnnotationType.label]

        with open(path, "r", encoding="utf-8") as f:
            lines = f.readlines()

        line_ids = [
            line_idx
            for line_idx, line in enumerate(lines)
            if ("/" in line or "\\" in line) and "." in line
        ]  # a heuristic for paths

        for line_idx in line_ids:
            image_path = lines[line_idx].strip()
            item_id = osp.splitext(image_path)[0]
            item_id = item_id.replace("\\", "/")

            image_path = osp.join(
                self._dataset_dir,
                WiderFacePath.SUBSET_DIR + self._subset,
                WiderFacePath.IMAGES_DIR,
                image_path,
            )

            annotations = []
            if "/" in item_id:
                label_name = item_id.split("/")[0]
                if "--" in label_name:
                    label_name = label_name.split("--")[1]
                if label_name != WiderFacePath.IMAGES_DIR_NO_LABEL:
                    label = label_categories.find(label_name)[0]
                    if label is not None:
                        annotations.append(Label(label=label))
                        self._ann_types.add(AnnotationType.label)
                item_id = item_id[len(item_id.split("/")[0]) + 1 :]

            items[item_id] = DatasetItem(
                id=item_id,
                subset=self._subset,
                media=Image.from_file(path=image_path),
                annotations=annotations,
            )

            try:
                bbox_count = int(lines[line_idx + 1])
            except ValueError:  # can be the next image
                continue
            except IndexError:  # the file can only contain names of images
                continue

            bbox_lines = lines[line_idx + 2 : line_idx + bbox_count + 2]
            for bbox in bbox_lines:
                bbox_list = bbox.split()
                if 4 <= len(bbox_list):
                    label = label_categories.find(WiderFacePath.DEFAULT_LABEL)[0]
                    if len(bbox_list) == 5 or len(bbox_list) == 11:
                        label_name = bbox_list[-1]
                        label = label_categories.find(label_name)[0]
                    if label is None and len(label_categories) == 0:
                        label_categories.add(WiderFacePath.DEFAULT_LABEL)
                        label = label_categories.find(WiderFacePath.DEFAULT_LABEL)[0]

                    attributes = {}
                    if 10 <= len(bbox_list):
                        i = 4
                        for attr in WiderFacePath.BBOX_ATTRIBUTES:
                            if bbox_list[i] != "-":
                                if bbox_list[i] in ["True", "False"]:
                                    attributes[attr] = str_to_bool(bbox_list[i])
                                else:
                                    attributes[attr] = bbox_list[i]
                            i += 1

                    items[item_id].annotations.append(
                        Bbox(
                            float(bbox_list[0]),
                            float(bbox_list[1]),
                            float(bbox_list[2]),
                            float(bbox_list[3]),
                            attributes=attributes,
                            label=label,
                        )
                    )
                    self._ann_types.add(AnnotationType.bbox)

        return items


class WiderFaceImporter(Importer):
    _ANNO_EXT = ".txt"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        context.require_file(f"{WiderFacePath.ANNOTATIONS_DIR}/*{cls._ANNO_EXT}")

    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(
            path, cls._ANNO_EXT, "wider_face", dirname=WiderFacePath.ANNOTATIONS_DIR
        )

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]


class WiderFaceExporter(Exporter):
    DEFAULT_IMAGE_EXT = WiderFacePath.IMAGE_EXT

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        save_dir = self._save_dir
        os.makedirs(save_dir, exist_ok=True)

        label_categories = self._extractor.categories()[AnnotationType.label]

        if self._save_dataset_meta:
            self._save_meta_file(save_dir)
        else:
            labels_path = osp.join(save_dir, WiderFacePath.LABELS_FILE)
            with open(labels_path, "w", encoding="utf-8") as f:
                f.write("\n".join(label.name for label in label_categories))

        for subset_name, subset in self._extractor.subsets().items():
            subset_dir = osp.join(save_dir, WiderFacePath.SUBSET_DIR + subset_name)

            wider_annotation = ""
            for item in subset:
                labels = [a.label for a in item.annotations if a.type == AnnotationType.label]
                if labels:
                    image_path = self._make_image_filename(
                        item, subdir="%s--%s" % (labels[0], label_categories[labels[0]].name)
                    )
                else:
                    image_path = self._make_image_filename(
                        item, subdir=WiderFacePath.IMAGES_DIR_NO_LABEL
                    )
                wider_annotation += image_path + "\n"
                if item.media and self._save_media:
                    self._save_image(
                        item, osp.join(save_dir, subset_dir, WiderFacePath.IMAGES_DIR, image_path)
                    )

                bboxes = [a for a in item.annotations if a.type == AnnotationType.bbox]
                if 0 < len(bboxes):
                    wider_annotation += "%s\n" % len(bboxes)
                for bbox in bboxes:
                    wider_bb = " ".join("%s" % p for p in bbox.get_bbox())
                    wider_annotation += "%s " % wider_bb
                    if bbox.attributes:
                        wider_attr = ""
                        attr_counter = 0
                        for attr in WiderFacePath.BBOX_ATTRIBUTES:
                            if attr in bbox.attributes:
                                wider_attr += "%s " % bbox.attributes[attr]
                                attr_counter += 1
                            else:
                                wider_attr += "- "
                        if 0 < attr_counter:
                            wider_annotation += wider_attr
                    if (
                        label_categories[bbox.label].name != WiderFacePath.DEFAULT_LABEL
                        and bbox.label is not None
                    ):
                        wider_annotation += "%s" % label_categories[bbox.label].name
                    wider_annotation += "\n"

            annotation_path = osp.join(
                save_dir, WiderFacePath.ANNOTATIONS_DIR, "wider_face_" + subset_name + "_bbx_gt.txt"
            )
            os.makedirs(osp.dirname(annotation_path), exist_ok=True)
            with open(annotation_path, "w", encoding="utf-8") as f:
                f.write(wider_annotation)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from typing import List, Tuple

import cv2
import numpy as np

from datumaro.components.abstracts import IModelInterpreter
from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred, PrepInfo
from datumaro.components.annotation import Annotation, AnnotationType, LabelCategories
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.errors import DatumaroError
from datumaro.components.media import Image
from datumaro.plugins.openvino_plugin.samples.utils import gen_hash_key
from datumaro.util.samples import get_samples_path


class ClipVisualViTB32ModelInterpreter(IModelInterpreter):
    mean = (255 * np.array([0.485, 0.456, 0.406])).reshape(1, 1, 3)
    std = (255 * np.array([0.229, 0.224, 0.225])).reshape(1, 1, 3)

    def preprocess(self, inp: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        img = inp.media_as(Image).data
        img = cv2.resize(img, (224, 224))
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = (img - self.mean) / self.std
        if len(img.shape) == 3 and img.shape[2] in {3, 4}:
            img = np.transpose(img, (2, 0, 1))
        return img, None

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        feature_vector = pred.get("output")
        if feature_vector is None:
            raise DatumaroError('"output" key should exist in the model prediction.')

        return [gen_hash_key(feature_vector)]

    def get_categories(self):
        label_categories = LabelCategories()

        openvino_plugin_samples_dir = get_samples_path()
        imagenet_class_path = osp.join(openvino_plugin_samples_dir, "imagenet.class")
        with open(imagenet_class_path, "r", encoding="utf-8") as file:
            for line in file.readlines():
                label = line.strip()
                label_categories.add(label)

        return {AnnotationType.label: label_categories}



# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import csv
import errno
import glob
import logging as log
import os.path as osp
from typing import List, Optional

import numpy as np

from datumaro.components.annotation import Bbox, Caption, ExtractedMask, MaskCategories, Polygon
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import InvalidAnnotationError
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.image import IMAGE_EXTENSIONS, find_images
from datumaro.util.mask_tools import lazy_mask

from .format import IcdarPath, IcdarTask


class _IcdarBase(SubsetBase):
    def __init__(
        self,
        path: str,
        task: IcdarTask,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        self._path = path
        self._task = task

        if task is IcdarTask.word_recognition:
            if not osp.isfile(path):
                raise FileNotFoundError(errno.ENOENT, "Can't find annotations file", path)

            if not subset:
                subset = osp.basename(osp.dirname(path))
            super().__init__(subset=subset, ctx=ctx)

            self._dataset_dir = osp.dirname(osp.dirname(path))

            self._items = list(self._load_recognition_items().values())
        elif task in {IcdarTask.text_localization, IcdarTask.text_segmentation}:
            if not osp.isdir(path):
                raise NotADirectoryError(
                    errno.ENOTDIR, "Can't read dataset directory with annotation files", path
                )

            if not subset:
                subset = osp.basename(path)
            super().__init__(subset=subset, ctx=ctx)

            self._dataset_dir = osp.dirname(path)

            if task is IcdarTask.text_localization:
                self._items = list(self._load_localization_items().values())
            else:
                self._items = list(self._load_segmentation_items().values())

    def _load_recognition_items(self):
        items = {}

        with open(self._path, encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                objects = line.split(", ")
                if len(objects) == 2:
                    image = objects[0]
                    captions = []
                    for caption in objects[1:]:
                        if caption[0] != '"' or caption[-1] != '"':
                            log.warning("Line %s: unexpected number " "of quotes" % line)
                        else:
                            captions.append(caption.replace("\\", "")[1:-1])
                else:
                    image = objects[0][:-1]
                    captions = []

                item_id = osp.splitext(image)[0]
                image_path = osp.join(osp.dirname(self._path), IcdarPath.IMAGES_DIR, image)
                if item_id not in items:
                    items[item_id] = DatasetItem(
                        item_id, subset=self._subset, media=Image.from_file(path=image_path)
                    )

                annotations = items[item_id].annotations
                for caption in captions:
                    annotations.append(Caption(caption))

                for ann in annotations:
                    self._ann_types.add(ann.type)

        return items

    def _load_localization_items(self):
        items = {}

        image_dir = osp.join(self._path, IcdarPath.IMAGES_DIR)
        if osp.isdir(image_dir):
            images = {
                osp.splitext(osp.relpath(p, image_dir))[0].replace("\\", "/"): p
                for p in find_images(image_dir, recursive=True)
            }
        else:
            images = {}

        for path in glob.iglob(osp.join(self._path, "**", "*.txt"), recursive=True):
            item_id = osp.splitext(osp.relpath(path, self._path))[0]
            if osp.basename(item_id).startswith("gt_"):
                item_id = osp.join(osp.dirname(item_id), osp.basename(item_id)[3:])
            item_id = item_id.replace("\\", "/")

            if item_id not in items:
                image = None
                image_path = images.get(item_id)
                if image_path:
                    image = Image.from_file(path=image_path)

                items[item_id] = DatasetItem(item_id, subset=self._subset, media=image)
            annotations = items[item_id].annotations

            with open(path, encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    objects = line.split('"')
                    if 1 < len(objects):
                        if len(objects) == 3:
                            text = objects[1]
                        else:
                            raise InvalidAnnotationError(
                                "Line %s: unexpected number " "of quotes in filename" % line
                            )
                    else:
                        text = ""
                    objects = objects[0].split()
                    if len(objects) == 1:
                        objects = objects[0].split(",")

                    if 8 <= len(objects):
                        points = [float(p) for p in objects[:8]]

                        attributes = {}
                        if 0 < len(text):
                            attributes["text"] = text
                        elif len(objects) == 9:
                            text = objects[8]
                            attributes["text"] = text

                        annotations.append(Polygon(points, attributes=attributes))
                    elif 4 <= len(objects):
                        x = float(objects[0])
                        y = float(objects[1])
                        w = float(objects[2]) - x
                        h = float(objects[3]) - y

                        attributes = {}
                        if 0 < len(text):
                            attributes["text"] = text
                        elif len(objects) == 5:
                            text = objects[4]
                            attributes["text"] = text

                        annotations.append(Bbox(x, y, w, h, attributes=attributes))

            for ann in annotations:
                self._ann_types.add(ann.type)

        return items

    def _load_segmentation_items(self):
        items = {}

        image_dir = osp.join(self._path, IcdarPath.IMAGES_DIR)
        if osp.isdir(image_dir):
            images = {
                osp.splitext(osp.relpath(p, image_dir))[0].replace("\\", "/"): p
                for p in find_images(image_dir, recursive=True)
            }
        else:
            images = {}

        for path in glob.iglob(osp.join(self._path, "**", "*.txt"), recursive=True):
            item_id = osp.splitext(osp.relpath(path, self._path))[0]
            item_id = item_id.replace("\\", "/")
            if item_id.endswith("_GT"):
                item_id = item_id[:-3]

            if item_id not in items:
                image = None
                image_path = images.get(item_id)
                if image_path:
                    image = Image.from_file(path=image_path)

                items[item_id] = DatasetItem(item_id, subset=self._subset, media=image)
            annotations = items[item_id].annotations

            colors = [(255, 255, 255)]
            chars = [""]
            centers = [0]
            groups = [0]
            group = 1
            number_in_group = 0
            with open(path, encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    if line == "":
                        if number_in_group == 1:
                            groups[len(groups) - 1] = 0
                        else:
                            group += 1
                        number_in_group = 0
                        continue

                    objects = line.split()
                    if objects[0][0] == "#":
                        objects[0] = objects[0][1:]
                        objects[9] = '" "'
                        objects.pop()
                    if len(objects) != 10:
                        raise InvalidAnnotationError(
                            "Line %s contains the wrong number "
                            'of arguments, e.g. \'241 73 144 1 4 0 3 1 4 "h"' % line
                        )

                    centers.append(objects[3] + " " + objects[4])
                    groups.append(group)
                    colors.append(tuple(int(o) for o in objects[:3]))
                    char = objects[9]
                    if char[0] == '"' and char[-1] == '"':
                        char = char[1:-1]
                    chars.append(char)
                    number_in_group += 1
            if number_in_group == 1:
                groups[len(groups) - 1] = 0

            mask_categories = MaskCategories({i: colors[i] for i in range(len(colors))})
            inverse_cls_colormap = mask_categories.inverse_colormap

            gt_path = osp.join(self._path, item_id + "_GT" + IcdarPath.GT_EXT)
            if osp.isfile(gt_path):
                # load mask through cache
                index_mask = lazy_mask(gt_path, inverse_cls_colormap)
                np_index_mask = index_mask()

                classes = np.unique(np_index_mask)
                for label_id in classes:
                    if label_id == 0:
                        continue
                    i = int(label_id)
                    annotations.append(
                        ExtractedMask(
                            index_mask=index_mask,
                            index=label_id,
                            group=groups[i],
                            attributes={
                                "index": i - 1,
                                "color": " ".join(str(p) for p in colors[i]),
                                "text": chars[i],
                                "center": centers[i],
                            },
                        )
                    )

            for ann in annotations:
                self._ann_types.add(ann.type)

        return items

    @staticmethod
    def _lazy_extract_mask(mask, c):
        return lambda: mask == c


class IcdarWordRecognitionBase(_IcdarBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = IcdarTask.word_recognition
        super().__init__(path, **kwargs)


class IcdarTextLocalizationBase(_IcdarBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = IcdarTask.text_localization
        super().__init__(path, **kwargs)


class IcdarTextSegmentationBase(_IcdarBase):
    def __init__(self, path, **kwargs):
        kwargs["task"] = IcdarTask.text_segmentation
        super().__init__(path, **kwargs)


class IcdarWordRecognitionImporter(Importer):
    _ANNO_EXT = ".txt"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        annot_path = context.require_file(f"*/gt{cls._ANNO_EXT}")

        with context.probe_text_file(
            annot_path,
            "must be a ICDAR-like annotation file",
        ) as f:
            reader = csv.reader(f, doublequote=False, escapechar="\\", skipinitialspace=True)
            fields = next(reader)
            if len(fields) != 2:
                raise Exception
            if osp.splitext(fields[0])[1] not in IMAGE_EXTENSIONS:
                raise Exception

    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(path, ".txt", "icdar_word_recognition")

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]


class IcdarTextLocalizationImporter(Importer):
    _ANNO_EXT = ".txt"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        context.require_file(f"**/gt_*{cls._ANNO_EXT}")

    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(path, "", "icdar_text_localization")

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]


class IcdarTextSegmentationImporter(Importer):
    _ANNO_EXT = ".txt"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        gt_txt_path = context.require_file(f"**/*_GT{cls._ANNO_EXT}")
        gt_bmp_path = osp.splitext(gt_txt_path)[0] + ".bmp"
        context.require_file(glob.escape(gt_bmp_path))

    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(path, "", "icdar_text_segmentation")

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import csv
import errno
import os
import os.path as osp
from typing import Dict, List, Optional, Union

from defusedxml import ElementTree

from datumaro.components.annotation import (
    Annotation,
    AnnotationType,
    Bbox,
    Label,
    LabelCategories,
    RotatedBbox,
)
from datumaro.components.dataset import DatasetItem
from datumaro.components.dataset_base import SubsetBase
from datumaro.components.errors import InvalidAnnotationError, UndeclaredLabelError
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image, ImageFromFile
from datumaro.plugins.data_formats.coco.base import _CocoBase
from datumaro.plugins.data_formats.coco.format import CocoImporterType, CocoTask
from datumaro.plugins.data_formats.voc.base import VocBase
from datumaro.plugins.data_formats.voc.format import VocImporterType, VocTask
from datumaro.plugins.data_formats.yolo.base import YoloUltralyticsBase
from datumaro.util import parse_json_file
from datumaro.util.image import IMAGE_EXTENSIONS, ImageMeta
from datumaro.util.os_util import find_files


class RoboflowCocoBase(_CocoBase):
    """
    Parses Roboflow COCO annotations written in the following format:
    https://cocodataset.org/#format-data
    """

    def __init__(
        self,
        path,
        *,
        subset: Optional[str] = None,
        stream: bool = False,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(
            path,
            task=CocoTask.instances,
            coco_importer_type=CocoImporterType.roboflow,
            subset=subset,
            stream=stream,
            ctx=ctx,
        )


class RoboflowVocBase(VocBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(
            path,
            task=VocTask.voc_detection,
            voc_importer_type=VocImporterType.roboflow,
            subset=subset,
            ctx=ctx,
        )

    def _load_categories(self, path):
        cats = set()
        ann_files = [file for file in os.listdir(path) if file.endswith(".xml")]
        for ann_file in ann_files:
            xml_file = osp.join(path, ann_file)

            root = ElementTree.parse(xml_file).getroot()

            if root.tag != "annotation":
                continue

            for object_elem in root.iterfind("object"):
                cat_name = self._parse_field(object_elem, "name")
                cats.add(cat_name)

        label_categories = LabelCategories()
        for _, cat in enumerate(sorted(cats)):
            label_categories.add(cat)

        categories = {AnnotationType.label: label_categories}

        return categories

    def _load_subset_list(self, path):
        return [os.path.splitext(file)[0] for file in os.listdir(path) if file.endswith(".xml")]


class RoboflowYoloBase(YoloUltralyticsBase):
    def __init__(
        self,
        config_path: str,
        image_info: Union[None, str, ImageMeta] = None,
        urls: Optional[List[str]] = None,
        **kwargs,
    ) -> None:
        super().__init__(config_path, image_info, urls, **kwargs)

    def _get_rootpath(self, config_path: str) -> str:
        return osp.dirname(config_path)

    def _load_img_files(self, rootpath: str) -> Dict:
        return {
            self._get_fname(img_file): img_file
            for img_file in find_files(rootpath, IMAGE_EXTENSIONS, recursive=True, max_depth=2)
            if osp.split(osp.relpath(osp.dirname(img_file), rootpath))[0] == self._subset
        }


class RoboflowYoloObbBase(RoboflowYoloBase):
    def __init__(
        self,
        config_path: str,
        image_info: Union[None, str, ImageMeta] = None,
        urls: Optional[List[str]] = None,
        **kwargs,
    ) -> None:
        super().__init__(config_path, image_info, urls, **kwargs)

    def _parse_annotations(
        self,
        anno_path: str,
        image: ImageFromFile,
        *,
        label_categories: LabelCategories,
    ) -> List[Annotation]:
        lines = []
        with open(anno_path, "r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if line:
                    lines.append(line)

        annotations = []
        for idx, line in enumerate(lines):
            parts = line.split()
            if len(parts) != 10:
                raise InvalidAnnotationError(
                    f"Unexpected field count {len(parts)} in the bbox description. "
                    "Expected 10 fields (x1, y1, x2, y2, x3, y3, x4, y4, label, 0)."
                )
            x1, y1, x2, y2, x3, y3, x4, y4, label_name = parts[:-1]

            label_name = self._parse_field(label_name, str, "bbox label name")
            label_id = label_categories.find(label_name)[0]

            if label_id is None:
                raise UndeclaredLabelError(str(label_id))

            x1 = self._parse_field(x1, float, "x1")
            y1 = self._parse_field(y1, float, "y1")
            x2 = self._parse_field(x2, float, "x2")
            y2 = self._parse_field(y2, float, "y2")
            x3 = self._parse_field(x3, float, "x3")
            y3 = self._parse_field(y3, float, "y3")
            x4 = self._parse_field(x4, float, "x4")
            y4 = self._parse_field(y4, float, "y4")
            annotations.append(
                RotatedBbox.from_rectangle(
                    points=[(x1, y1), (x2, y2), (x3, y3), (x4, y4)],
                    label=label_id,
                    id=idx,
                    group=idx,
                )
            )

        return annotations


class RoboflowCreateMlBase(SubsetBase):
    """
    Parses Roboflow CreateML annotations written in the following format:
    https://cocodataset.org/#format-data
    """

    def __init__(
        self,
        path,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isfile(path):
            raise FileNotFoundError(errno.ENOENT, "Can't find JSON file", path)
        self._path = path

        if not subset:
            subset = osp.splitext(osp.basename(path))[0]
        super().__init__(subset=subset, ctx=ctx)

        json_data = parse_json_file(path)
        self._categories = self._load_categories(json_data)
        self._items = self._load_items(json_data)

    def _load_categories(self, json_data):
        cats = set()
        for anns in json_data:
            for ann in anns["annotations"]:
                cats.add(ann["label"])

        label_categories = LabelCategories()
        for cat in sorted(cats):
            label_categories.add(cat)

        categories = {AnnotationType.label: label_categories}

        return categories

    def _load_items(self, json_data):
        items = {}
        for anns in self._ctx.progress_reporter.iter(
            json_data, desc=f"Parsing boxes in '{self._subset}'"
        ):
            annotations = []
            for ann_id, ann in enumerate(anns["annotations"]):
                label_id, _ = self._categories[AnnotationType.label].find(ann["label"])
                if label_id is None:
                    raise UndeclaredLabelError(ann["label"])

                x = ann["coordinates"]["x"]
                y = ann["coordinates"]["y"]
                w = ann["coordinates"]["width"]
                h = ann["coordinates"]["height"]
                annotations.append(Bbox(x, y, w, h, label=label_id, id=ann_id, group=ann_id))
                self._ann_types.add(AnnotationType.bbox)

            img_id = osp.splitext(anns["image"])[0]
            items[img_id] = DatasetItem(
                id=img_id,
                subset=self._subset,
                media=Image.from_file(path=osp.join(osp.dirname(self._path), anns["image"])),
                annotations=annotations,
            )

        return items.values()


class RoboflowMulticlassBase(SubsetBase):
    """
    Parses Roboflow Multiclass annotations written in the following format:
    https://cocodataset.org/#format-data
    """

    def __init__(
        self,
        path,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isfile(path):
            raise FileNotFoundError(errno.ENOENT, "Can't find CSV file", path)
        self._path = path

        if not subset:
            subset = osp.splitext(osp.basename(path))[0]
        super().__init__(subset=subset, ctx=ctx)

        self._label_mapping = {}
        self._categories = self._load_categories(path)
        self._items = self._load_items(path)

    def _load_categories(self, path):
        with open(path, "r", encoding="utf-8") as f:
            reader = csv.DictReader(f)
            cats = [label.strip() for label in reader.fieldnames[1:]]

        label_categories = LabelCategories()
        for idx, cat in enumerate(sorted(cats)):
            label_categories.add(cat)
            self._label_mapping[cat] = idx

        categories = {AnnotationType.label: label_categories}

        return categories

    def _load_items(self, path):
        items = []
        with open(path, "r", encoding="utf-8") as f:
            for anns in csv.DictReader(f):
                img_id = anns.get("filename", None)
                img_id = osp.splitext(img_id)[0] if img_id else None
                idx = 0
                annotations = []
                for key, val in anns.items():
                    if key.strip() not in self._label_mapping:
                        continue
                    if int(val) == 1:
                        annotations.append(
                            Label(label=self._label_mapping[key.strip()], id=idx, group=idx)
                        )
                        self._ann_types.add(AnnotationType.label)
                        idx += 1

                items.append(
                    DatasetItem(
                        id=img_id,
                        subset=self._subset,
                        media=Image.from_file(
                            path=osp.join(osp.dirname(self._path), anns["filename"])
                        ),
                        annotations=annotations,
                    )
                )

        return items



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
import tempfile
from multiprocessing.pool import Pool
from shutil import move, rmtree
from typing import Any, Callable, Dict, Iterator, Optional, Union

import numpy as np
import pyarrow as pa

from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.errors import DatumaroError
from datumaro.components.exporter import ExportContext, Exporter
from datumaro.util.multi_procs_util import consumer_generator

from .format import DatumaroArrow
from .mapper.dataset_item import DatasetItemMapper
from .mapper.media import ImageMapper


class ArrowExporter(Exporter):
    AVAILABLE_IMAGE_EXTS = ImageMapper.AVAILABLE_SCHEMES
    DEFAULT_IMAGE_EXT = ImageMapper.AVAILABLE_SCHEMES[0]

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)

        # '--image-ext' would be used in a different way for arrow foramt
        _actions = []
        for action in parser._actions:
            if action.dest != "image_ext":
                _actions.append(action)
        parser._actions = _actions
        parser._option_string_actions.pop("--image-ext")

        parser.add_argument(
            "--image-ext",
            default=None,
            help=f"Image encoding scheme. (default: {cls.DEFAULT_IMAGE_EXT})",
            choices=cls.AVAILABLE_IMAGE_EXTS,
        )

        parser.add_argument(
            "--max-shard-size",
            type=int,
            default=1000,
            help="The maximum number of dataset item can be stored in each shard file. "
            "'--max-shard-size' and '--num-shards' are mutually exclusive, "
            "Therefore, if '--max-shard-size' is not None, the number of shard files will be determined by "
            "(# of total dataset item) / (max chunk size). "
            "(default: %(default)s)",
        )

        parser.add_argument(
            "--num-shards",
            type=int,
            default=None,
            help="The number of shards to export. "
            "'--max-shard-size' and '--num-shards' are mutually exclusive. "
            "Therefore, if '--num-shards' is not None, the number of dataset item in each shard file "
            "will be determined by (# of total dataset item) / (num shards). "
            "(default: %(default)s)",
        )

        parser.add_argument(
            "--num-workers",
            type=int,
            default=0,
            help="The number of multi-processing workers for export. "
            "If num_workers = 0, do not use multiprocessing. (default: %(default)s)",
        )

        parser.add_argument(
            "--prefix",
            type=str,
            default="datum",
            help="Prefix to be appended in front of the shard file name. "
            "Therefore, the generated file name will be `<prefix>-<idx>.arrow`. "
            "(default: %(default)s)",
        )

        return parser

    def _apply_impl(self, *args, **kwargs):
        if self._num_workers == 0:
            return self._apply()

        with Pool(processes=self._num_workers) as pool:
            return self._apply(pool)

    def _apply(self, pool: Optional[Pool] = None):
        os.makedirs(self._save_dir, exist_ok=True)

        pbar = self._ctx.progress_reporter

        if pool is not None:

            def _producer_gen():
                for item in self._extractor:
                    future = pool.apply_async(
                        func=self._item_to_dict_record,
                        args=(item, self._image_ext, self._source_path),
                    )
                    yield future

            with consumer_generator(producer_generator=_producer_gen()) as consumer_gen:

                def _gen_with_pbar():
                    for item in pbar.iter(
                        consumer_gen, desc="Exporting", total=len(self._extractor)
                    ):
                        yield item.get()

                self._write_file(_gen_with_pbar())

        else:

            def create_consumer_gen():
                for item in pbar.iter(self._extractor, desc="Exporting"):
                    yield self._item_to_dict_record(item, self._image_ext, self._source_path)

            self._write_file(create_consumer_gen())

    def _write_file(self, consumer_gen: Iterator[Dict[str, Any]]) -> None:
        for file_idx, size in enumerate(self._chunk_sizes):
            suffix = str(file_idx).zfill(self._max_digits)
            fname = f"{self._prefix}-{suffix}.arrow"
            fpath = os.path.join(self._save_dir, fname)

            record_batch = pa.RecordBatch.from_pylist(
                mapping=[next(consumer_gen) for _ in range(size)],
                schema=self._schema,
            )

            with pa.OSFile(fpath, "wb") as sink:
                with pa.ipc.new_file(sink, self._schema) as writer:
                    writer.write(record_batch)

        pass

    @classmethod
    def patch(cls, dataset, patch, save_dir, **kwargs):
        # no patch supported
        with tempfile.TemporaryDirectory() as temp_dir:
            cls.convert(dataset, save_dir=temp_dir, **kwargs)
            if os.path.exists(save_dir):
                for file in os.listdir(save_dir):
                    file = os.path.join(save_dir, file)
                    if os.path.isdir(file):
                        rmtree(file)
                    else:
                        os.remove(file)
            for file in os.listdir(temp_dir):
                file_from = os.path.join(temp_dir, file)
                file_to = os.path.join(save_dir, file)
                move(file_from, file_to)

    def __init__(
        self,
        extractor: IDataset,
        save_dir: str,
        *,
        save_media: Optional[bool] = None,
        image_ext: Optional[Union[str, Callable[[str], bytes]]] = None,
        default_image_ext: Optional[str] = None,
        save_dataset_meta: bool = False,
        ctx: Optional[ExportContext] = None,
        num_workers: int = 0,
        max_shard_size: Optional[int] = 1000,
        num_shards: Optional[int] = None,
        prefix: str = "datum",
        **kwargs,
    ):
        super().__init__(
            extractor=extractor,
            save_dir=save_dir,
            save_media=save_media,
            image_ext=image_ext,
            default_image_ext=default_image_ext,
            save_dataset_meta=save_dataset_meta,
            ctx=ctx,
        )

        if num_workers < 0:
            raise DatumaroError(
                f"num_workers should be non-negative but num_workers={num_workers}."
            )
        self._num_workers = num_workers

        if num_shards is not None and max_shard_size is not None:
            raise DatumaroError(
                "Both 'num_shards' or 'max_shard_size' cannot be provided at the same time."
            )
        elif num_shards is not None and num_shards < 0:
            raise DatumaroError(f"num_shards should be non-negative but num_shards={num_shards}.")
        elif max_shard_size is not None and max_shard_size < 0:
            raise DatumaroError(
                f"max_shard_size should be non-negative but max_shard_size={max_shard_size}."
            )
        elif num_shards is None and max_shard_size is None:
            raise DatumaroError(
                "Either one of 'num_shards' or 'max_shard_size' should be provided."
            )

        self._num_shards = num_shards
        self._max_shard_size = max_shard_size

        if self._save_media:
            self._image_ext = (
                self._image_ext if self._image_ext is not None else self._default_image_ext
            )
        else:
            self._image_ext = "NONE"

        assert (
            self._image_ext in self.AVAILABLE_IMAGE_EXTS
        ), f"{self._image_ext} is unkonwn ext. Available exts are {self.AVAILABLE_IMAGE_EXTS}"

        self._prefix = prefix

        self._source_path = (
            os.path.abspath(self._extractor._source_path)
            if getattr(self._extractor, "_source_path")
            else None
        )

        total_len = len(self._extractor)

        if self._num_shards is not None:
            max_shard_size = int(total_len / self._num_shards) + 1

        elif self._max_shard_size is not None:
            max_shard_size = self._max_shard_size
        else:
            raise DatumaroError(
                "Either one of 'num_shards' or 'max_shard_size' should be provided."
            )

        self._chunk_sizes = np.diff(
            np.array([size for size in range(0, total_len, max_shard_size)] + [total_len])
        )
        assert (
            sum(self._chunk_sizes) == total_len
        ), "Sum of chunk sizes should be the number of total items."

        num_shard_files = len(self._chunk_sizes)
        self._max_digits = len(str(num_shard_files))

        self._schema = DatumaroArrow.create_schema_with_metadata(self._extractor)

        self._subsets = {
            subset_name: idx for idx, subset_name in enumerate(self._extractor.subsets())
        }

    @staticmethod
    def _item_to_dict_record(
        item: DatasetItem,
        image_ext: Optional[str] = None,
        source_path: Optional[str] = None,
    ) -> Dict[str, Any]:
        dict_item = DatasetItemMapper.forward(item, media={"encoder": image_ext})

        def _change_path(parent: Dict) -> Dict:
            for key, child in parent.items():
                if key == "path" and child is not None:
                    parent["path"] = child.replace(source_path, "")

                if isinstance(child, dict):
                    parent[key] = _change_path(child)

            return parent

        if source_path is not None:
            return _change_path(dict_item)

        return dict_item



# Copyright (C) 2021-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import os
import os.path as osp
from enum import Enum
from typing import List, Optional
from zipfile import ZIP_BZIP2, ZIP_DEFLATED, ZIP_LZMA, ZIP_STORED, ZipFile

from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.exporter import Exporter
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util import parse_str_enum_value
from datumaro.util.image import IMAGE_EXTENSIONS, encode_image


class Compression(Enum):
    ZIP_STORED = ZIP_STORED
    ZIP_DEFLATED = ZIP_DEFLATED
    ZIP_BZIP2 = ZIP_BZIP2
    ZIP_LZMA = ZIP_LZMA


class ImageZipPath:
    DEFAULT_ARCHIVE_NAME = "default.zip"
    DEFAULT_COMPRESSION = Compression.ZIP_STORED


class ImageZipBase(SubsetBase):
    def __init__(
        self,
        url: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        super().__init__(subset=subset, media_type=Image, ctx=ctx)

        assert url.endswith(".zip"), url

        with ZipFile(url, "r") as zf:
            for path in zf.filelist:
                item_id, extension = osp.splitext(path.filename)
                if extension.lower() not in IMAGE_EXTENSIONS:
                    continue
                image = Image.from_bytes(data=zf.read(path.filename))
                self._items.append(DatasetItem(id=item_id, media=image, subset=self._subset))

        self._ann_types = set()


class ImageZipImporter(Importer):
    _FORMAT_EXT = ".zip"

    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(path, cls._FORMAT_EXT, "image_zip")

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._FORMAT_EXT]


class ImageZipExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".jpg"

    @staticmethod
    def _get_compression_method(s):
        try:
            return Compression[s.upper()]
        except KeyError:
            import argparse

            raise argparse.ArgumentTypeError()

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)

        parser.add_argument(
            "--name",
            type=str,
            default=ImageZipPath.DEFAULT_ARCHIVE_NAME,
            help="Name of output zipfile (default: %(default)s)",
        )

        parser.add_argument(
            "--compression",
            type=cls._get_compression_method,
            default=ImageZipPath.DEFAULT_COMPRESSION.name,
            help="Archive compression method.\nAvailable methods: {} "
            "(default: %(default)s)".format(", ".join(e.name for e in Compression)),
        )

        return parser

    def __init__(self, extractor, save_dir, name=None, compression=None, **kwargs):
        super().__init__(extractor, save_dir, **kwargs)

        if name is None:
            name = ImageZipPath.DEFAULT_ARCHIVE_NAME

        compression = parse_str_enum_value(
            compression, Compression, default=ImageZipPath.DEFAULT_COMPRESSION
        )

        self._archive_name = name
        self._compression = compression.value

    def _apply_impl(self):
        os.makedirs(self._save_dir, exist_ok=True)

        archive_path = osp.join(self._save_dir, self._archive_name)

        if osp.exists(archive_path):
            raise FileExistsError(
                "Zip file: %s, already exist, "
                "specify archive name with --name extra argument" % archive_path
            )

        with ZipFile(archive_path, "w", self._compression) as zf:
            for item in self._extractor:
                if item.media:
                    self._archive_image(zf, item)
                else:
                    log.debug("Item '%s' has no image info", item.id)

    def _archive_image(self, zipfile, item):
        image_name = self._make_image_filename(item)
        path = getattr(item.media, "path", None)
        if path is not None and osp.isfile(path):
            zipfile.write(item.media.path, arcname=image_name)
        elif item.media.has_data:
            zipfile.writestr(image_name, encode_image(item.media.data, osp.splitext(image_name)[1]))



# Copyright (C) 2023 Intel Corporation
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# We implemented some of this code by referring to the codebase available at
# https://github.com/facebookresearch/segment-anything.
# This code is licensed under the Apache License 2.0, which can be found in the LICENSE file at
# https://github.com/facebookresearch/segment-anything/blob/main/LICENSE.

from typing import List, Tuple

import numpy as np
from attr import attrs, field

from datumaro.components.abstracts import IModelInterpreter
from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred, PrepInfo
from datumaro.components.annotation import Annotation, FeatureVector, Mask
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.media import Image
from datumaro.util.annotation_util import nms


@attrs(slots=True, kw_only=True, order=False)
class AMGPoints(Annotation):
    """Intermediate annotation class for SAM decoder inputs.

    Attributes:
        points: Array of points (x, y) for the SAM prompt.
    """

    points: np.ndarray = field()


@attrs(slots=True, kw_only=True, order=False)
class AMGMasks(Annotation):
    """Intermediate annotation class for SAM decoder outputs.

    Attributes:
        masks: Array of masks corresponded to the points.
        iou_preds: Array of Intersection over Union (IoU)
            prediction scores corresponded to the points.
    """

    masks: np.ndarray = field()
    iou_preds: np.ndarray = field()

    @classmethod
    def cat(cls, masks: List["AMGMasks"]) -> "AMGMasks":
        """Concatenate a list of `AMGMasks` into a single `AMGMasks` object.

        Parameters:
            masks: List of `AMGMasks` to concatenate.

        Returns:
            A new AMGMasks containing the concatenated masks and IoU prediction scores.
        """
        return AMGMasks(
            masks=np.concatenate([mask.masks for mask in masks], axis=0),
            iou_preds=np.concatenate([mask.iou_preds for mask in masks], axis=0),
        )

    def postprocess(
        self,
        mask_threshold: float,
        pred_iou_thresh: float,
        stability_score_offset: float,
        stability_score_thresh: float,
        box_nms_thresh: float,
        min_mask_region_area: int,
    ) -> List[Mask]:
        """Postprocesses the masks with the given parameters.

        Parameters:
            pred_iou_thresh (float): A filtering threshold in [0,1], using the
                model's predicted mask quality.
            stability_score_thresh (float): A filtering threshold in [0,1], using
                the stability of the mask under changes to the cutoff used to binarize
                the model's mask predictions.
            stability_score_offset (float): The amount to shift the cutoff when
                calculated the stability score.
            box_nms_thresh (float): The box IoU cutoff used by non-maximal
                suppression to filter duplicate masks.
            min_mask_region_area (int): If >0, postprocessing will be applied
                to remove the binary mask which has the number of 1s less than min_mask_region_area.

        Returns:
            List of :class:`Mask`s representing the postprocessed masks.
        """
        masks, iou_preds = self.masks, self.iou_preds

        if pred_iou_thresh > 0.0:
            keep_mask = iou_preds > pred_iou_thresh
            masks = masks[keep_mask]
            iou_preds = iou_preds[keep_mask]

        if stability_score_thresh > 0.0:
            keep_mask = (
                self._calculate_stability_score(
                    masks=masks,
                    mask_threshold=mask_threshold,
                    stability_score_offset=stability_score_offset,
                )
                > stability_score_thresh
            )
            masks = masks[keep_mask]
            iou_preds = iou_preds[keep_mask]

        binary_masks = masks > mask_threshold

        if min_mask_region_area > 0:
            keep_mask = binary_masks.sum(axis=(1, 2)) > min_mask_region_area
            binary_masks = binary_masks[keep_mask]
            iou_preds = iou_preds[keep_mask]

        segments = [
            Mask(
                image=binary_mask,
                id=idx,
                group=idx,
                object_id=idx,
                label=0,
                z_order=0,
                attributes={"score": iou_pred},
            )
            for idx, (binary_mask, iou_pred) in enumerate(zip(binary_masks, iou_preds))
        ]
        segments_after_nms = nms(segments=segments, iou_thresh=box_nms_thresh)
        return segments_after_nms

    @staticmethod
    def _calculate_stability_score(
        masks: np.ndarray, mask_threshold: float, stability_score_offset: float
    ) -> np.ndarray:
        """
        Computes the stability score for a batch of masks. The stability
        score is the IoU between the binary masks obtained by thresholding
        the predicted mask logits at high and low values.
        """
        # One mask is always contained inside the other.
        # Save memory by preventing unnecessary cast to torch.int64
        intersections = (
            (masks > (mask_threshold + stability_score_offset))
            .sum(-1, dtype=np.int16)
            .sum(-1, dtype=np.int32)
        )
        unions = (
            (masks > (mask_threshold - stability_score_offset))
            .sum(-1, dtype=np.int16)
            .sum(-1, dtype=np.int32)
        )
        return intersections / unions


class SAMDecoderForAMGInterpreter(IModelInterpreter):
    """Interpreter for the automatic mask generation using SAM decoder."""

    h_model = 1024
    w_model = 1024
    onnx_mask_input = np.zeros((1, 1, 256, 256), dtype=np.float32)
    onnx_has_mask_input = np.zeros(1, dtype=np.float32)

    def preprocess(self, inp: DatasetItem) -> Tuple[LauncherInputType, PrepInfo]:
        img_size = inp.media_as(Image).size

        img_embed = inp.annotations[-1]

        assert isinstance(
            img_embed, FeatureVector
        ), "annotations should have the image embedding vector as FeatureVector."

        amg_points = inp.annotations[-2]

        h_img, w_img = img_size

        points = amg_points.points
        points[:, 0] *= self.w_model
        points[:, 1] *= self.h_model

        scale = min(self.h_model / h_img, self.w_model / w_img)

        if h_img <= w_img:
            points[:, 1] *= scale
        else:
            points[:, 0] *= scale

        onnx_coord = np.concatenate(
            [points.reshape(-1, 1, 2), np.zeros_like(points).reshape(-1, 1, 2)], axis=1
        ).astype(np.float32)
        onnx_label = np.array([1, -1] * len(points)).reshape(-1, 2).astype(np.float32)

        decoder_inputs = {
            "image_embeddings": img_embed.vector[None, :, :, :],
            "point_coords": onnx_coord,
            "point_labels": onnx_label,
            "mask_input": self.onnx_mask_input,
            "has_mask_input": self.onnx_has_mask_input,
            "orig_im_size": np.array(img_size, dtype=np.float32),
        }

        return decoder_inputs, None

    def postprocess(self, pred: ModelPred, info: PrepInfo) -> List[Annotation]:
        """Postprocesses the outputs of the SAM decoder to generate masks automatically
        from the prompts which have a point uniformly distributed on a 2d grid.

        Parameters:
            pred: List of dictionaries containing model predictions. Each dictionary should have the 'masks'
            and 'iou_preds' keys. 'masks' is corresponding to the predicted mask of which shape is (1, H, W).
            'iou_preds' is corresponding to the scalar IoU prediction score.
            info: None

        Returns:
            List of :class:`AMGMasks` produced by the SAM decoder.
        """
        masks = np.concatenate([p["masks"] for p in pred], axis=0)
        iou_preds = np.concatenate([p["iou_predictions"] for p in pred], axis=0)
        return [AMGMasks(masks=masks, iou_preds=iou_preds)]

    def get_categories(self):
        return None



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.cli.util import MultilineFormatter, add_subparser
from datumaro.cli.util.errors import CliException
from datumaro.cli.util.project import generate_next_file_name
from datumaro.components.dataset import Dataset


def build_split_video_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Split video into frames",
        description="""
        Splits a video into separate frames and saves them in a directory.
        After the splitting, the images can be added into a project
        using the 'import' command and the 'image_dir' format.|n
        |n
        This command is useful for making a dataset from a video file.
        Unlike direct video reading during model training, which can produce
        different results if the system environment changes, this command
        allows to split the video into frames and use them instead, making
        the dataset reproducible and stable.|n
        |n
        This command provides different options like setting the frame step,
        file name pattern, starting and finishing frame etc.|n
        |n
        Examples:|n
        - Split a video into frames, use each 30-rd frame:|n
        |s|s%(prog)s -i video.mp4 -o video.mp4-frames --step 30|n
        - Split a video into frames, save as 'frame_xxxxxx.png' files:|n
        |s|s%(prog)s -i video.mp4 --image-ext=.png --name-pattern='frame_%%06d'
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "-i", "--input-path", dest="src_path", required=True, help="Path to the video file"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="Directory to save output (default: a subdir in the current one)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument(
        "-n",
        "--name-pattern",
        default="%06d",
        help="Name pattern for the produced images (default: %(default)s)",
    )
    parser.add_argument(
        "-s", "--step", type=int, default=1, help="Frame step (default: %(default)s)"
    )
    parser.add_argument(
        "-b", "--start-frame", type=int, default=0, help="Starting frame (default: %(default)s)"
    )
    parser.add_argument(
        "-e", "--end-frame", type=int, default=None, help="Finishing frame (default: %(default)s)"
    )
    parser.add_argument(
        "-x", "--image-ext", default=".jpg", help="Output image extension (default: %(default)s)"
    )
    parser.set_defaults(command=split_video_command)

    return parser


def get_split_video_sensitive_args():
    return {
        split_video_command: ["src_path", "dst_dir", "name_pattern"],
    }


def split_video_command(args):
    src_path = osp.abspath(args.src_path)

    dst_dir = args.dst_dir
    if dst_dir:
        if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
            raise CliException(
                "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
            )
    else:
        dst_dir = generate_next_file_name("%s-frames" % osp.basename(src_path))
    dst_dir = osp.abspath(dst_dir)

    log.info("Exporting frames...")

    dataset = Dataset.import_from(
        src_path,
        "video_frames",
        name_pattern=args.name_pattern,
        step=args.step,
        start_frame=args.start_frame,
        end_frame=args.end_frame,
    )

    dataset.export(format="image_dir", save_dir=dst_dir, image_ext=args.image_ext)

    log.info("Frames are exported into '%s'" % dst_dir)

    return 0


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor()

    subparsers = parser.add_subparsers()
    add_subparser(subparsers, "split_video", build_split_video_parser)

    return parser


def get_sensitive_args():
    return {
        **get_split_video_sensitive_args(),
    }



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from collections import OrderedDict
from typing import Dict, Sequence

import attr
from attr import attrib, attrs

from datumaro.components.annotation import (
    AnnotationType,
    LabelCategories,
    MaskCategories,
    PointsCategories,
)
from datumaro.components.annotations.merger import (
    AnnotationMerger,
    BboxMerger,
    CaptionsMerger,
    Cuboid2DMerger,
    Cuboid3dMerger,
    EllipseMerger,
    FeatureVectorMerger,
    HashKeyMerger,
    ImageAnnotationMerger,
    LabelMerger,
    LineMerger,
    MaskMerger,
    PointsMerger,
    PolygonMerger,
    RotatedBboxMerger,
    TabularMerger,
)
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.dataset_item_storage import (
    DatasetItemStorage,
    DatasetItemStorageDatasetView,
)
from datumaro.components.errors import (
    AnnotationsTooCloseError,
    ConflictingCategoriesError,
    FailedAttrVotingError,
    NoMatchingAnnError,
    NoMatchingItemError,
    WrongGroupError,
)
from datumaro.components.merge import Merger
from datumaro.util import find
from datumaro.util.annotation_util import find_instances, max_bbox
from datumaro.util.attrs_util import ensure_cls

__all__ = ["IntersectMerge"]


@attrs
class IntersectMerge(Merger):
    """
    Merge several datasets with "intersect" policy:

    - If there are two or more dataset items whose (id, subset) pairs match each other,
    we can consider this as having an intersection in our dataset. This method merges
    the annotations of the corresponding :class:`DatasetItem` into one :class:`DatasetItem`
    to handle this intersection. The rule to handle merging annotations is provided by
    :class:`AnnotationMerger` according to their annotation types. For example,
    DatasetItem(id="item_1", subset="train", annotations=[Bbox(0, 0, 1, 1)]) from Dataset-A and
    DatasetItem(id="item_1", subset="train", annotations=[Bbox(.5, .5, 1, 1)]) from Dataset-B can be
    merged into DatasetItem(id="item_1", subset="train", annotations=[Bbox(0, 0, 1, 1)]).

    - Label categories are merged according to the union of their label names
    (Same as `UnionMerge`). For example, if Dataset-A has {"car", "cat", "dog"}
    and Dataset-B has {"car", "bus", "truck"} labels, the merged dataset will have
    {"bust", "car", "cat", "dog", "truck"} labels.

    - This merge has configuration parameters (`conf`) to control the annotation merge behaviors.

    For example,

    ```python
    merge = IntersectMerge(
        conf=IntersectMerge.Conf(
            pairwise_dist=0.25,
            groups=[],
            output_conf_thresh=0.0,
            quorum=0,
        )
    )
    ```

    For more details for the parameters, please refer to :class:`IntersectMerge.Conf`.
    """

    def __init__(self, **options):
        super().__init__(**options)

    @attrs(repr_ns="IntersectMerge", kw_only=True)
    class Conf:
        """
        Parameters
        ----------
        pairwise_dist
            IoU match threshold for segments
        sigma
            Parameter for Object Keypoint Similarity metric
            (https://cocodataset.org/#keypoints-eval)
        output_conf_thresh
            Confidence threshold for output annotations
        quorum
            Minimum count for a label and attribute voting results to be counted
        ignored_attributes
            Attributes to be ignored in the merged :class:`DatasetItem`
        groups
            A comma-separated list of labels in annotation groups to check.
            '?' postfix can be added to a label to make it optional in the group (repeatable)
        close_distance
            Distance threshold between annotations to decide their closeness. If they are decided
            to be close, it will be enrolled to the error tracker.
        """

        pairwise_dist = attrib(converter=float, default=0.5)
        sigma = attrib(converter=list, factory=list)

        output_conf_thresh = attrib(converter=float, default=0)
        quorum = attrib(converter=int, default=0)
        ignored_attributes = attrib(converter=set, factory=set)

        def _groups_converter(value):
            result = []
            for group in value:
                rg = set()
                for label in group:
                    optional = label.endswith("?")
                    name = label if not optional else label[:-1]
                    rg.add((name, optional))
                result.append(rg)
            return result

        groups = attrib(converter=_groups_converter, factory=list)
        close_distance = attrib(converter=float, default=0.75)

    conf = attrib(converter=ensure_cls(Conf), factory=Conf)

    # Error trackers:
    errors = attrib(factory=list, init=False)

    def add_item_error(self, error, *args, **kwargs):
        self.errors.append(error(self._item_id, *args, **kwargs))

    # Indexes:
    _dataset_map = attrib(init=False)  # id(dataset) -> (dataset, index)
    _item_map = attrib(init=False)  # id(item) -> (item, id(dataset))
    _ann_map = attrib(init=False)  # id(ann) -> (ann, id(item))
    _item_id = attrib(init=False)
    _item = attrib(init=False)

    # Misc.
    _infos = attrib(init=False)  # merged infos
    _categories = attrib(init=False)  # merged categories

    def merge(self, sources: Sequence[IDataset]) -> DatasetItemStorage:
        self._infos = self.merge_infos([d.infos() for d in sources])
        self._categories = self.merge_categories([d.categories() for d in sources])
        merged = DatasetItemStorage()
        self._check_groups_definition()

        item_matches, item_map = self.match_items(sources)
        self._item_map = item_map
        self._dataset_map = {id(d): (d, i) for i, d in enumerate(sources)}

        for item_id, items in item_matches.items():
            self._item_id = item_id

            if len(items) < len(sources):
                missing_sources = set(id(s) for s in sources) - set(items)
                missing_sources = [self._dataset_map[s][1] for s in missing_sources]
                self.add_item_error(NoMatchingItemError, sources=missing_sources)
            merged.put(self.merge_items(items))

        return merged

    def get_ann_source(self, ann_id):
        return self._item_map[self._ann_map[ann_id][1]][1]

    def __call__(self, *datasets: IDataset) -> DatasetItemStorageDatasetView:
        # TODO: self.merge() should be the first since this order matters for
        # IntersectMerge.
        merged = self.merge(datasets)
        infos = self.merge_infos(d.infos() for d in datasets)
        categories = self.merge_categories(d.categories() for d in datasets)
        media_type = self.merge_media_types(datasets)
        ann_types = self.merge_ann_types(datasets)
        return DatasetItemStorageDatasetView(
            parent=merged,
            infos=infos,
            categories=categories,
            media_type=media_type,
            ann_types=ann_types,
        )

    def merge_categories(self, sources: Sequence[IDataset]) -> Dict:
        # TODO: This is a temporary workaround to minimize code changes.
        # We have to revisit it to make this class stateless.
        if hasattr(self, "_categories"):
            return self._categories

        dst_categories = {}

        label_cat = self._merge_label_categories(sources)
        if label_cat is None:
            label_cat = LabelCategories()
        dst_categories[AnnotationType.label] = label_cat

        points_cat = self._merge_point_categories(sources, label_cat)
        if points_cat is not None:
            dst_categories[AnnotationType.points] = points_cat

        mask_cat = self._merge_mask_categories(sources, label_cat)
        if mask_cat is not None:
            dst_categories[AnnotationType.mask] = mask_cat

        return dst_categories

    def merge_items(self, items: Dict[int, DatasetItem]) -> DatasetItem:
        self._item = next(iter(items.values()))

        self._ann_map = {}
        sources = []
        for item in items.values():
            self._ann_map.update({id(a): (a, id(item)) for a in item.annotations})
            sources.append(item.annotations)
        log.debug(
            "Merging item %s: source annotations %s" % (self._item_id, list(map(len, sources)))
        )

        annotations = self.merge_annotations(sources)

        annotations = [
            a for a in annotations if self.conf.output_conf_thresh <= a.attributes.get("score", 1)
        ]

        return self._item.wrap(annotations=annotations)

    def merge_annotations(self, sources):
        self._make_mergers(sources)

        clusters = self._match_annotations(sources)

        joined_clusters = sum(clusters.values(), [])
        group_map = self._find_cluster_groups(joined_clusters)

        annotations = []
        for t, clusters in clusters.items():
            for cluster in clusters:
                self._check_cluster_sources(cluster)

            merged_clusters = self._merge_clusters(t, clusters)

            for merged_ann, cluster in zip(merged_clusters, clusters):
                attributes = self._find_cluster_attrs(cluster, merged_ann)
                attributes = {
                    k: v for k, v in attributes.items() if k not in self.conf.ignored_attributes
                }
                attributes.update(merged_ann.attributes)
                merged_ann.attributes = attributes

                new_group_id = find(enumerate(group_map), lambda e: id(cluster) in e[1][0])
                if new_group_id is None:
                    new_group_id = 0
                else:
                    new_group_id = new_group_id[0] + 1
                merged_ann.group = new_group_id

            if self.conf.close_distance:
                self._check_annotation_distance(t, merged_clusters)

            annotations += merged_clusters

        if self.conf.groups:
            self._check_groups(annotations)

        return annotations

    def match_items(self, datasets):
        item_ids = set((item.id, item.subset) for d in datasets for item in d)

        item_map = {}  # id(item) -> (item, id(dataset))

        matches = OrderedDict()
        for item_id, item_subset in sorted(item_ids, key=lambda e: e[0]):
            items = {}
            for d in datasets:
                item = d.get(item_id, subset=item_subset)
                if item:
                    items[id(d)] = item
                    item_map[id(item)] = (item, id(d))
            matches[(item_id, item_subset)] = items

        return matches, item_map

    def _merge_label_categories(self, sources):
        same = True
        common = None
        for src_categories in sources:
            src_cat = src_categories.get(AnnotationType.label)
            if common is None:
                common = src_cat
            elif common != src_cat:
                same = False
                break

        if same:
            return common

        dst_cat = LabelCategories()
        for src_id, src_categories in enumerate(sources):
            src_cat = src_categories.get(AnnotationType.label)
            if src_cat is None:
                continue

            for src_label in src_cat.items:
                dst_label = dst_cat.find(src_label.name)[1]
                if dst_label is not None:
                    if dst_label != src_label:
                        if (
                            src_label.parent
                            and dst_label.parent
                            and src_label.parent != dst_label.parent
                        ):
                            raise ConflictingCategoriesError(
                                "Can't merge label category %s (from #%s): "
                                "parent label conflict: %s vs. %s"
                                % (src_label.name, src_id, src_label.parent, dst_label.parent),
                                sources=list(range(src_id)),
                            )
                        dst_label.parent = dst_label.parent or src_label.parent
                        dst_label.attributes |= src_label.attributes
                    else:
                        pass
                else:
                    dst_cat.add(src_label.name, src_label.parent, src_label.attributes)

        return dst_cat

    def _merge_point_categories(self, sources, label_cat):
        dst_point_cat = PointsCategories()

        for src_id, src_categories in enumerate(sources):
            src_label_cat = src_categories.get(AnnotationType.label)
            src_point_cat = src_categories.get(AnnotationType.points)
            if src_label_cat is None or src_point_cat is None:
                continue

            for src_label_id, src_cat in src_point_cat.items.items():
                src_label = src_label_cat.items[src_label_id].name
                dst_label_id = label_cat.find(src_label)[0]
                dst_cat = dst_point_cat.items.get(dst_label_id)
                if dst_cat is not None:
                    if dst_cat != src_cat:
                        raise ConflictingCategoriesError(
                            "Can't merge point category for label "
                            "%s (from #%s): %s vs. %s" % (src_label, src_id, src_cat, dst_cat),
                            sources=list(range(src_id)),
                        )
                    else:
                        pass
                else:
                    dst_point_cat.add(dst_label_id, src_cat.labels, src_cat.joints)

        if len(dst_point_cat.items) == 0:
            return None

        return dst_point_cat

    def _merge_mask_categories(self, sources, label_cat):
        dst_mask_cat = MaskCategories()

        for src_id, src_categories in enumerate(sources):
            src_label_cat = src_categories.get(AnnotationType.label)
            src_mask_cat = src_categories.get(AnnotationType.mask)
            if src_label_cat is None or src_mask_cat is None:
                continue

            for src_label_id, src_cat in src_mask_cat.colormap.items():
                src_label = src_label_cat.items[src_label_id].name
                dst_label_id = label_cat.find(src_label)[0]
                dst_cat = dst_mask_cat.colormap.get(dst_label_id)
                if dst_cat is not None:
                    if dst_cat != src_cat:
                        raise ConflictingCategoriesError(
                            "Can't merge mask category for label "
                            "%s (from #%s): %s vs. %s" % (src_label, src_id, src_cat, dst_cat),
                            sources=list(range(src_id)),
                        )
                    else:
                        pass
                else:
                    dst_mask_cat.colormap[dst_label_id] = src_cat

        if len(dst_mask_cat.colormap) == 0:
            return None

        return dst_mask_cat

    def _match_annotations(self, sources):
        all_by_type = {}
        for s in sources:
            src_by_type = {}
            for a in s:
                src_by_type.setdefault(a.type, []).append(a)
            for k, v in src_by_type.items():
                all_by_type.setdefault(k, []).append(v)

        clusters = {}
        for k, v in all_by_type.items():
            clusters.setdefault(k, []).extend(self._match_ann_type(k, v))

        return clusters

    def _make_mergers(self, sources):
        def _make(c, **kwargs):
            kwargs.update(attr.asdict(self.conf))
            fields = attr.fields_dict(c)
            return c(**{k: v for k, v in kwargs.items() if k in fields}, context=self)

        def _for_type(t, **kwargs):
            if t is AnnotationType.unknown:
                return _make(AnnotationMerger, **kwargs)
            elif t is AnnotationType.label:
                return _make(LabelMerger, **kwargs)
            elif t is AnnotationType.bbox:
                return _make(BboxMerger, **kwargs)
            elif t is AnnotationType.mask:
                return _make(MaskMerger, **kwargs)
            elif t is AnnotationType.polygon:
                return _make(PolygonMerger, **kwargs)
            elif t is AnnotationType.polyline:
                return _make(LineMerger, **kwargs)
            elif t is AnnotationType.points:
                return _make(PointsMerger, **kwargs)
            elif t is AnnotationType.caption:
                return _make(CaptionsMerger, **kwargs)
            elif t is AnnotationType.cuboid_3d:
                return _make(Cuboid3dMerger, **kwargs)
            elif t is AnnotationType.super_resolution_annotation:
                return _make(ImageAnnotationMerger, **kwargs)
            elif t is AnnotationType.depth_annotation:
                return _make(ImageAnnotationMerger, **kwargs)
            elif t is AnnotationType.ellipse:
                return _make(EllipseMerger, **kwargs)
            elif t is AnnotationType.hash_key:
                return _make(HashKeyMerger, **kwargs)
            elif t is AnnotationType.feature_vector:
                return _make(FeatureVectorMerger, **kwargs)
            elif t is AnnotationType.tabular:
                return _make(TabularMerger, **kwargs)
            elif t is AnnotationType.rotated_bbox:
                return _make(RotatedBboxMerger, **kwargs)
            elif t is AnnotationType.cuboid_2d:
                return _make(Cuboid2DMerger, **kwargs)
            else:
                raise NotImplementedError("Type %s is not supported" % t)

        instance_map = {}
        for s in sources:
            s_instances = find_instances(s)
            for inst in s_instances:
                inst_bbox = max_bbox(
                    [
                        a
                        for a in inst
                        if a.type
                        in {AnnotationType.polygon, AnnotationType.mask, AnnotationType.bbox}
                    ]
                )
                for ann in inst:
                    instance_map[id(ann)] = [inst, inst_bbox]

        self._mergers = {t: _for_type(t, instance_map=instance_map) for t in AnnotationType}

    def _match_ann_type(self, t, sources):
        return self._mergers[t].match_annotations(sources)

    def _merge_clusters(self, t, clusters):
        return self._mergers[t].merge_clusters(clusters)

    def _find_cluster_groups(self, clusters):
        cluster_groups = []
        visited = set()
        for a_idx, cluster_a in enumerate(clusters):
            if a_idx in visited:
                continue
            visited.add(a_idx)

            cluster_group = {id(cluster_a)}

            # find segment groups in the cluster group
            a_groups = set(ann.group for ann in cluster_a)
            for cluster_b in clusters[a_idx + 1 :]:
                b_groups = set(ann.group for ann in cluster_b)
                if a_groups & b_groups:
                    a_groups |= b_groups

            # now we know all the segment groups in this cluster group
            # so we can find adjacent clusters
            for b_idx, cluster_b in enumerate(clusters[a_idx + 1 :]):
                b_idx = a_idx + 1 + b_idx
                b_groups = set(ann.group for ann in cluster_b)
                if a_groups & b_groups:
                    cluster_group.add(id(cluster_b))
                    visited.add(b_idx)

            if a_groups == {0}:
                continue  # skip annotations without a group
            cluster_groups.append((cluster_group, a_groups))
        return cluster_groups

    def _find_cluster_attrs(self, cluster, ann):
        quorum = self.conf.quorum or 0

        # TODO: when attribute types are implemented, add linear
        # interpolation for contiguous values

        attr_votes = {}  # name -> { value: score , ... }
        for s in cluster:
            for name, value in s.attributes.items():
                votes = attr_votes.get(name, {})
                votes[value] = 1 + votes.get(value, 0)
                attr_votes[name] = votes

        attributes = {}
        for name, votes in attr_votes.items():
            winner, count = max(votes.items(), key=lambda e: e[1])
            if count < quorum:
                if sum(votes.values()) < quorum:
                    # blame provokers
                    missing_sources = set(
                        self.get_ann_source(id(a))
                        for a in cluster
                        if s.attributes.get(name) == winner
                    )
                else:
                    # blame outliers
                    missing_sources = set(
                        self.get_ann_source(id(a))
                        for a in cluster
                        if s.attributes.get(name) != winner
                    )
                missing_sources = [self._dataset_map[s][1] for s in missing_sources]
                self.add_item_error(
                    FailedAttrVotingError, name, votes, ann, sources=missing_sources
                )
                continue
            attributes[name] = winner

        return attributes

    def _check_cluster_sources(self, cluster):
        if len(cluster) == len(self._dataset_map):
            return

        def _has_item(s):
            item = self._dataset_map[s][0].get(*self._item_id)
            if not item:
                return False
            if len(item.annotations) == 0:
                return False
            return True

        missing_sources = set(self._dataset_map) - set(self.get_ann_source(id(a)) for a in cluster)
        missing_sources = [self._dataset_map[s][1] for s in missing_sources if _has_item(s)]
        if missing_sources:
            self.add_item_error(NoMatchingAnnError, cluster[0], sources=missing_sources)

    def _check_annotation_distance(self, t, annotations):
        for a_idx, a_ann in enumerate(annotations):
            for b_ann in annotations[a_idx + 1 :]:
                d = self._mergers[t].distance(a_ann, b_ann)
                if self.conf.close_distance < d:
                    self.add_item_error(AnnotationsTooCloseError, a_ann, b_ann, d)

    def _check_groups(self, annotations):
        check_groups = []
        for check_group_raw in self.conf.groups:
            check_group = set(l[0] for l in check_group_raw)
            optional = set(l[0] for l in check_group_raw if l[1])
            check_groups.append((check_group, optional))

        def _check_group(group_labels, group):
            for check_group, optional in check_groups:
                common = check_group & group_labels
                real_miss = check_group - common - optional
                extra = group_labels - check_group
                if common and (extra or real_miss):
                    self.add_item_error(WrongGroupError, group_labels, check_group, group)
                    break

        groups = find_instances(annotations)
        for group in groups:
            group_labels = set()
            for ann in group:
                if not hasattr(ann, "label"):
                    continue
                label = self._get_label_name(ann.label)

                if ann.group:
                    group_labels.add(label)
                else:
                    _check_group({label}, [ann])

            if not group_labels:
                continue
            _check_group(group_labels, group)

    def _get_label_name(self, label_id):
        if label_id is None:
            return None
        return self._categories[AnnotationType.label].items[label_id].name

    def _get_label_id(self, label):
        return self._categories[AnnotationType.label].find(label)[0]

    def _get_src_label_name(self, ann, label_id):
        if label_id is None:
            return None
        item_id = self._ann_map[id(ann)][1]
        dataset_id = self._item_map[item_id][1]
        return (
            self._dataset_map[dataset_id][0].categories()[AnnotationType.label].items[label_id].name
        )

    def get_any_label_name(self, ann, label_id):
        if label_id is None:
            return None
        try:
            return self._get_src_label_name(ann, label_id)
        except KeyError:
            return self._get_label_name(label_id)

    def _check_groups_definition(self):
        for group in self.conf.groups:
            for label, _ in group:
                _, entry = self._categories[AnnotationType.label].find(label)
                if entry is None:
                    raise ValueError(
                        "Datasets do not contain "
                        "label '%s', available labels %s"
                        % (label, [i.name for i in self._categories[AnnotationType.label].items])
                    )
